TABLE_HEADER = (
    'DPS', 'Combat Time', 'Combat Time Share', 'Total Damage', 'Debuff', 'Attacks-in Share',
    'Taken Damage Share', 'Damage Share', 'Max One Hit', 'Crit Chance', 'Deaths', 'Total Heals',
    'Heal Share', 'Heal Crit Chance', 'Total Damage Taken', 'Total Hull Damage Taken',
    'Total Shield Damage Taken', 'Total Attacks', 'Hull Attacks', 'Attacks-in Number',
    'Heal Crit Number', 'Heal Number', 'Crit Number', 'Misses')

TREE_HEADER = (
    '', 'DPS', 'Total Damage', 'Debuff', 'Max One Hit', 'Crit Chance', 'Accuracy', 'Flank Rate',
    'Kills', 'Attacks', 'Misses', 'Critical Hits', 'Flank Hits', 'Shield Damage', 'Shield DPS',
    'Hull Damage', 'Hull DPS', 'Base Damage', 'Base DPS', 'Combat Time', 'Hull Attacks',
    'Shield Attacks')  # , 'Shield Resistance')

HEAL_TREE_HEADER = (
    '', 'HPS', 'Total Heal', 'Hull Heal', 'Hull HPS', 'Shield Heal', 'Shield HPS',
    'Max One Heal', 'Crit Chance', 'Heal Ticks', 'Critical Heals', 'Combat Time', 'Hull Heal Ticks',
    'Shield Heal Ticks')

LIVE_TABLE_HEADER = (
    'DPS', 'Combat Time', 'Debuff', 'Attacks-in', 'HPS', 'Kills', 'Deaths')

BANNED_ABILITIES = {
    'Electrical Overload'}

PATCHES = (
    (b'Rehona, Sister of the Qowat Milat', b'Rehona - Sister of the Qowat Milat'),
    (b'"Pull Together, Fall Apart"', b'Pull Together - Fall Apart')
)

# tuple(string in the first line identifying the issue, text with linebreaks removed, replacement
#       text (replaces text with linebreaks removed), total number of lines the issue spans)
MULTILINE_PATCHES = (
    (b'"Nanite Infection',
     b"Nanite Infection<br>Causes damage to nearby players and Kobayashi Maru",
     b'"Nanite Infection - Causes damage to nearby players and Kobayashi Maru"', 3),
)



from copy import deepcopy
import os
import time
from threading import Event, Lock, Thread, Timer
from types import FunctionType, BuiltinFunctionType, MethodType

from .utilities import to_datetime

CALLABLE = (FunctionType, BuiltinFunctionType, MethodType)


def _f(*args, **kwargs):
    return


class LiveParser():
    '''
    OSCR's realtime parser
    '''

    def __init__(self, log_path, start_callback=None, update_callback=None, settings: dict = None):
        '''
        Creates LiveParser Instance.

        Parameters:
        - :param log_path: path to combatlog that is being analyzed
        - :param start_callback: function that is called immediately before the live parser starts
        to analyze the logfile initiated by the LiveParser.start() function; no arguments will be
        passed
        - :param update_callback: function that is called once every second when the parser is
        running; it will be passed a dictionary with the player data acquired in the last second as
        positional argument and the total combat time as second positional argument
        - :param settings: contains settings
            - "seconds_between_combats": number of inactive seconds after which a new engagement
            resets the collected data
        '''
        if not (os.path.exists(log_path) and os.path.isfile(log_path)):
            raise FileNotFoundError('Invalid Logpath')
        self._active = Event()
        self._lock = Lock()
        self._players = dict()
        self._inactive_seconds = 0
        self._reset = False
        if isinstance(start_callback, CALLABLE):
            self.start_callback = start_callback
        else:
            self.start_callback = _f
        if isinstance(update_callback, CALLABLE):
            self.update_callback = update_callback
        else:
            self.update_callback = _f
        self._update_timer = Timer(interval=1, function=self.update_data)
        self.log_path = log_path
        self.settings = {
            'seconds_between_combats': 100
        }
        for key in self.settings:
            if key in settings:
                self.settings[key] = settings[key]

    def __del__(self):
        """
        Stops all active threads when the object is garbage-collected.
        """
        self.stop()

    def start(self):
        """
        Initiates the parsing process.
        """
        if not self._active.is_set():
            self.start_callback()
            analyzer_thread = Thread(target=self.analyze, args=())
            analyzer_thread.start()
            self._update_timer = Timer(interval=1, function=self.update_data)
            self._update_timer.start()

    def stop(self):
        """
        Terminates the parsing process. Already running threads might perform one more data
        iteration before the process ends.
        """
        self._active.clear()
        self._update_timer.cancel()

    def update_data(self):
        """
        Refines the data and executes the update_callback. Runs once per second when active.
        """
        if not self._active:
            return
        self._update_timer = Timer(interval=1, function=self.update_data)
        self._update_timer.start()
        if len(self._players) < 1:
            return
        total_attacks_in = 0
        with self._lock:
            player_copy = deepcopy(self._players)
            for player in self._players.values():
                total_attacks_in += player['attacks_in_buffer']
                player['attacks_in_buffer'] = 0
                player['base_damage_buffer'] = 0
                player['damage_buffer'] = 0
        output = dict()
        first_player_attacks = list()
        last_player_attacks = list()
        for player, player_data in player_copy.items():
            _, handle_part = player.split(' ', maxsplit=1)
            index = handle_part.index('@')
            name_and_handle = (handle_part[:index], handle_part[index:-1])
            if player_data['combat_start'] is not None:
                first_player_attacks.append(player_data['combat_start'])
                last_player_attacks.append(player_data['combat_end'])
                combat_time = player_data['combat_end'] - player_data['combat_start']
                try:
                    dps = player_data['damage'] / combat_time
                except ZeroDivisionError:
                    dps = 0
                try:
                    hps = player_data['heal'] / combat_time
                except ZeroDivisionError:
                    hps = 0
            else:
                dps = 0
                hps = 0
                combat_time = 0
            try:
                debuff = (player_data['damage_buffer'] / player_data['base_damage_buffer']) - 1
            except ZeroDivisionError:
                debuff = 0
            try:
                attacks_in_share = player_data['attacks_in_buffer'] / total_attacks_in
            except ZeroDivisionError:
                attacks_in_share = 0
            output[name_and_handle] = {
                'dps': dps,
                'combat_time': combat_time,
                'local_debuff': debuff * 100,
                'local_attacks_in_share': attacks_in_share * 100,
                'hps': hps,
                'kills': player_data['kills'],
                'deaths': player_data['deaths']
            }
        if len(first_player_attacks) > 0:
            player_combat_duration = max(last_player_attacks) - min(first_player_attacks)
        else:
            player_combat_duration = 0
        self.update_callback(output, player_combat_duration)

    def analyze(self):
        """
        Analyzes the log continuously until LiveParser.stop() is called. Clears existing data first
        when called.
        """
        with self._lock:
            self._players = dict()
        self._reset = False
        with open(self.log_path, 'r', encoding='utf-8') as logfile:
            logfile.seek(0, 2)
            self._active.set()
            while self._active.is_set():
                line = logfile.readline()
                if not line:
                    time.sleep(0.5)
                    if self._reset:
                        continue
                    elif self._inactive_seconds >= self.settings['seconds_between_combats']:
                        self._inactive_seconds = 0
                        self._reset = True
                    else:
                        self._inactive_seconds += 0.5
                    continue
                if self._reset:
                    with self._lock:
                        self._players = dict()
                    self._reset = False
                self._inactive_seconds = 0
                time_data, attack_data = line.split('::')
                timestamp = to_datetime(time_data).timestamp()
                attack_data = attack_data.split(',')
                player_attacks = attack_data[1].startswith("P")
                player_attacked = attack_data[5].startswith("P") and not attack_data[2]
                if not player_attacks and not player_attacked:
                    continue
                magnitude = float(attack_data[10])
                magnitude2 = float(attack_data[11])
                is_shield = attack_data[8] == 'Shield'
                is_heal = (
                        (is_shield and magnitude < 0 and magnitude2 >= 0)
                        or (attack_data[8] == 'HitPoints' and magnitude < 0))
                is_kill = 'Kill' in attack_data[9]
                magnitude = abs(magnitude)
                magnitude2 = abs(magnitude2)

                attacker_id = attack_data[1]
                target_id = attack_data[5]

                if player_attacks:
                    if attacker_id not in self._players:
                        with self._lock:
                            self._players[attacker_id] = {
                                'damage': 0,
                                'combat_start': None,
                                'combat_end': None,
                                'base_damage_buffer': 0,
                                'damage_buffer': 0,
                                'heal': 0,
                                'attacks_in_buffer': 0,
                                'kills': 0,
                                'deaths': 0
                            }
                            if not is_heal and attack_data[5] != '*':
                                self._players[attacker_id]['combat_start'] = timestamp
                    if attack_data[3] == '*' and attack_data[5] == '*':
                        pass
                    elif is_heal:
                        with self._lock:
                            self._players[attacker_id]['heal'] += magnitude
                    else:
                        if self._players[attacker_id]['combat_start'] is None:
                            with self._lock:
                                self._players[attacker_id]['combat_start'] = timestamp
                        with self._lock:
                            self._players[attacker_id]['combat_end'] = timestamp
                            self._players[attacker_id]['damage'] += magnitude
                            self._players[attacker_id]['damage_buffer'] += magnitude
                            self._players[attacker_id]['base_damage_buffer'] += magnitude2
                            if is_kill:
                                self._players[attacker_id]['kills'] += 1
                if player_attacked and not is_shield:
                    if target_id not in self._players:
                        with self._lock:
                            self._players[target_id] = {
                                'damage': 0,
                                'combat_start': None,
                                'base_damage_buffer': 0,
                                'damage_buffer': 0,
                                'heal': 0,
                                'attacks_in_buffer': 0,
                                'kills': 0,
                                'deaths': 0
                            }
                    with self._lock:
                        self._players[target_id]['attacks_in_buffer'] += 1
                        if is_kill:
                            self._players[target_id]['deaths'] += 1



import gzip
import io
import os

_81920 = io.DEFAULT_BUFFER_SIZE * 10


class ReadFileBackwards():
    """reads an utf-8 encoded text file and yields its lines backwards in a memory efficient way"""

    __slots__ = (
            '_buffer_size', '_file', '_path', '_offset', 'filesize', '_position', '_remainder',
            '_lines', '_iter_counter')

    def __init__(self, path: str, offset: int = 0, buffer_size: int = _81920):
        """
        Reads utf-8 encoded text file.

        Parameters:
        - :param path: path to the text file
        - :param offset: number of bytes to ignore from the end of the file
        - :param buffer_size: number of bytes to buffer
        """
        self._buffer_size = buffer_size
        self._file = None
        self._path = path
        self._offset = offset
        self.filesize = 0
        self._position = -1
        self._remainder = bytes()
        self._lines = None
        self._iter_counter = None

    @property
    def top(self):
        """next line, None if there is no next line"""
        try:
            return self._lines[-1 - self._iter_counter]
        except IndexError:
            return None

    @property
    def total_bytes_read(self):
        """
        Number of bytes read. -1 if file has not been opened yet or is still open.
        offset not included
        """
        if self._file is not None and self._file.closed:
            return self.filesize - self._position - self._offset
        else:
            return -1

    def open(self):
        """Openes the file. (usually done by the context manager)"""
        self.__enter__()

    def close(self):
        """Closes the file. (usually done by the context manager)"""
        self.__exit__()

    def get_bytes_read(self, ignore_last_line: bool = False):
        """
        Bytes read excluding offset and number of lines.
        """
        ignore_lines = 1 if ignore_last_line else 0
        not_consumed_bytes = self._calculate_not_consumed_bytes(ignore_lines)
        return self.filesize - self._position - not_consumed_bytes - self._offset

    def __enter__(self):
        self._file = open(self._path, 'rb')
        if self._file.read(2) == b'\x1f\x8b':
            self._file.close()
            self._file = gzip.open(self._path, 'rb')
        self._file.seek(0, os.SEEK_END)
        self.filesize = self._file.tell()
        self._position = self._file.seek(self.filesize - self._offset)
        self._lines = self._get_chunk()
        self._iter_counter = 0
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        self._file.close()
        if self._position > 0:
            self._position += self._calculate_not_consumed_bytes()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self._iter_counter += 1
            return self._lines[-self._iter_counter]
        except IndexError:
            next_chunk = self._get_chunk()
            if len(next_chunk) == 0:
                self._iter_counter = -1
                raise StopIteration()
            else:
                self._lines = next_chunk
                self._iter_counter = 1
                return self._lines[-1]

    def _get_chunk(self):
        new_position = self._position - self._buffer_size
        if new_position <= 0:
            self._file.seek(0, 0)
            # self._position contains the number of bytes *before* it, so reading that many bytes
            # returns everything from the beginning up to (not including) the byte at self.position
            new_text = (self._file.read(self._position) + self._remainder).decode('utf-8').strip()
            new_lines = new_text.splitlines(keepends=True)
            self._remainder = bytes()
            self._position = 0
        else:
            self._position = self._file.seek(new_position, 0)
            new_bytes = self._file.read(self._buffer_size) + self._remainder
            try:
                self._remainder, line_bytes = new_bytes.split(b'\n', 1)
                self._remainder += b'\n'
            except ValueError:
                self._remainder = bytes()
                line_bytes = new_bytes
            new_lines = line_bytes.decode('utf-8').splitlines(keepends=True)
        return new_lines

    def _calculate_not_consumed_bytes(self, ignore_lines: int = 0):
        if self._iter_counter == -1:
            return 0
        not_consumed = ''.join(self._lines[:ignore_lines - self._iter_counter])
        return len(self._remainder + not_consumed.encode('utf-8'))



import numpy
from datetime import datetime

from .combat import Combat
from .constants import HEAL_TREE_HEADER, TREE_HEADER
from .datamodels import (
    AnalysisTableRow, DamageTableRow, HealTableRow, LogLine, TreeItem, TreeModel)
from .utilities import bundle, get_handle_from_id, get_player_handle, to_microseconds


def analyze_combat(combat: Combat) -> Combat:
    """
    Fully analyzes the given combat and returns it.
    """
    combat.damage_out = dmg_out_model = TreeModel(TREE_HEADER)
    combat.damage_in = dmg_in_model = TreeModel(TREE_HEADER)
    combat.heals_out = heal_out_model = TreeModel(HEAL_TREE_HEADER)
    combat.heals_in = heal_in_model = TreeModel(HEAL_TREE_HEADER)
    actor_combat_durations: dict[str, list[datetime]] = dict()
    graph_point_delta = combat.graph_resolution * 1_000_000
    combat_duration_delta = combat.end_time - combat.start_time
    combat_duration_sec = int(combat_duration_delta.total_seconds()) + 1  # round up to full second
    total_graph_points = int(combat_duration_delta.total_seconds() // combat.graph_resolution + 2)
    combat_start = combat.log_data[0].timestamp
    relative_combat_sec = 0
    for line in combat.log_data:
        timestamp = line.timestamp
        player_attacks = line.owner_id.startswith('P')
        player_attacked = line.target_id.startswith('P')
        is_shield_line = line.type == 'Shield'
        crit_flag, miss_flag, flank_flag, kill_flag, _, _ = get_flags(line.flags)
        is_heal = (
                (line.type == 'HitPoints' and line.magnitude < 0)
                or (is_shield_line and line.magnitude < 0 and line.magnitude2 >= 0))

        relative_combat_sec = (timestamp - combat_start).seconds

        # HEALS
        if is_heal:
            target_item, ability_target = get_outgoing_target_row(
                    heal_out_model, line, player_attacks, HealTableRow, combat_duration_sec)
            source_item, source_ability = get_incoming_target_row(
                    heal_in_model, line, player_attacked, HealTableRow, combat_duration_sec)

            if crit_flag:
                ability_target.critical_heals += 1
                source_ability.critical_heals += 1

            magnitude = abs(line.magnitude)
            ability_target.total_heal += magnitude
            ability_target.heal_ticks += 1
            source_ability.total_heal += magnitude
            source_ability.heal_ticks += 1
            if is_shield_line:
                ability_target.shield_heal += magnitude
                ability_target.shield_heal_ticks += 1
                source_ability.shield_heal += magnitude
                source_ability.shield_heal_ticks += 1
            else:
                ability_target.hull_heal += magnitude
                ability_target.hull_heal_ticks += 1
                source_ability.hull_heal += magnitude
                source_ability.hull_heal_ticks += 1

            if magnitude > ability_target.max_one_heal:
                ability_target.max_one_heal = magnitude
            if magnitude > source_ability.max_one_heal:
                source_ability.max_one_heal = magnitude

            target_item.graph_data[relative_combat_sec] += magnitude
            source_item.graph_data[relative_combat_sec] += magnitude

        # DAMAGE
        else:
            target_item, ability_target = get_outgoing_target_row(
                    dmg_out_model, line, player_attacks, DamageTableRow, combat_duration_sec)
            source_item, source_ability = get_incoming_target_row(
                    dmg_in_model, line, player_attacked, DamageTableRow, combat_duration_sec)

            # Combat Duration
            # Heals, damage taken and self-damage don't affect combat time
            if ability_target.name != '*' and line.owner_id != line.target_id:
                try:
                    actor_combat_durations[line.owner_id][1] = timestamp
                except KeyError:
                    actor_combat_durations[line.owner_id] = [timestamp, timestamp]
                if not line.source_id.startswith('P'):
                    try:
                        actor_combat_durations[line.source_id][1] = timestamp
                    except KeyError:
                        actor_combat_durations[line.source_id] = [timestamp, timestamp]

            # get table data
            magnitude = abs(line.magnitude)
            magnitude2 = abs(line.magnitude2)
            ability_target.total_damage += magnitude
            ability_target.total_attacks += 1
            source_ability.total_damage += magnitude
            source_ability.total_attacks += 1
            if is_shield_line:
                ability_target.total_shield_damage += magnitude
                ability_target.shield_attacks += 1
                source_ability.total_shield_damage += magnitude
                source_ability.shield_attacks += 1
            else:
                ability_target.total_hull_damage += magnitude
                source_ability.total_hull_damage += magnitude
                ability_target.hull_attacks += 1
                source_ability.hull_attacks += 1
                ability_target.total_base_damage += magnitude2
                source_ability.total_base_damage += magnitude2

            if magnitude > ability_target.max_one_hit:
                ability_target.max_one_hit = magnitude
            if magnitude > source_ability.max_one_hit:
                source_ability.max_one_hit = magnitude

            target_item.graph_data[relative_combat_sec] += magnitude
            source_item.graph_data[relative_combat_sec] += magnitude

            # overview graph data
            if player_attacks:
                time_idx = int(to_microseconds(timestamp - combat_start) // graph_point_delta)
                try:
                    combat.overview_graphs[get_player_handle(line.owner_id)][time_idx] += magnitude
                except KeyError:
                    combat.overview_graphs[get_player_handle(line.owner_id)] = numpy.zeros(
                            total_graph_points, numpy.float64)
                    combat.overview_graphs[get_player_handle(line.owner_id)][time_idx] += magnitude

            if miss_flag:
                ability_target.misses += 1
                source_ability.misses += 1
            if flank_flag:
                ability_target.flank_num += 1
                source_ability.flank_num += 1
            if crit_flag:
                ability_target.crit_num += 1
                source_ability.crit_num += 1
            if kill_flag:
                ability_target.kills += 1
                source_ability.kills += 1
                if (line.target_name == 'Borg Queen Octahedron'
                        or (line.target_id == '*' and (line.owner_name == 'Borg Queen Octahedron'
                            or line.source_name == 'Borg Queen Octahedron'))):
                    if combat.map_is_hive_space():
                        combat_duration_delta = line.timestamp - combat.log_data[0].timestamp
                        combat.end_time = line.timestamp
                        break  # ignore all lines after the Queen kill line in the Hive Space queue

    combat.meta['log_duration'] = combat_duration_delta.total_seconds()
    overview_graph_intervals: dict[str, tuple] = dict()
    first_player_shot = list()
    last_player_shot = list()
    for actor_id, (start_time, end_time) in actor_combat_durations.items():
        if actor_id.startswith('P'):
            start = int((start_time - combat.start_time).total_seconds() // combat.graph_resolution)
            end = int((end_time - combat.start_time).total_seconds() // combat.graph_resolution + 1)
            overview_graph_intervals[get_player_handle(actor_id)] = (start, end)
            first_player_shot.append(start_time)
            last_player_shot.append(end_time)
        actor_combat_durations[actor_id] = round((end_time - start_time).total_seconds(), 1)
    if len(first_player_shot) > 0 and len(last_player_shot) > 0:
        combat.meta['player_duration'] = (
                max(last_player_shot) - min(first_player_shot)).total_seconds()
    else:
        combat.meta['player_duration'] = 0

    merge_single_lines(dmg_out_model)
    combat_duration = combat_duration_delta.total_seconds()
    complete_damage_tree(dmg_out_model, actor_combat_durations, combat_duration)
    complete_damage_tree(dmg_in_model, actor_combat_durations, combat_duration)
    complete_heal_tree(heal_out_model, actor_combat_durations, combat_duration)
    complete_heal_tree(heal_in_model, actor_combat_durations, combat_duration)
    combat.detect_map()
    combat.create_overview(overview_graph_intervals)
    return combat


def get_outgoing_target_row(
        tree_model: TreeModel, line: LogLine, player_attacks: bool, row_constructor,
        parse_duration: int) -> tuple[TreeItem, DamageTableRow | HealTableRow]:
    """
    Adds the needed parents to the tree model and returns the newly created or already existing
    data row. Also updates combat time of the actor.

    Parameters:
    - :param tree_model: model to work on
    - :param line: log line that contains the ability
    - :param player_attacks: True when player attacks, False otherwise
    - :param parse_duration: seconds between the first and last line of the combat, rounded up

    :return: reference to newly created or existing data row
    """
    attacker_id = (line.owner_id,)
    attacker_handle = get_handle_from_id(line.owner_id)
    if attacker_id in tree_model.actor_index:
        attacker = tree_model.actor_index[attacker_id]
    else:
        attacker = tree_model.add_actor(
                line.owner_name, attacker_handle, attacker_id, DamageTableRow, player_attacks)
        attacker.data.combat_start = line.timestamp
    attacker.data.combat_end = line.timestamp

    if line.source_name:
        pet_group_id = (line.owner_id, line.source_name)
        attacker_id = (line.owner_id, line.source_id)
        if pet_group_id in tree_model.pet_group_index:
            attacker = tree_model.pet_group_index[pet_group_id]
        # if attacking pet already in pet index: logfile is bugged (one pet has more than one
        # entity name); pet group is not created in this case as it would not have any children
        elif attacker_id not in tree_model.pet_index:
            attacker = tree_model.add_pet_group(line.source_name, pet_group_id, attacker)

        if attacker_id in tree_model.pet_index:
            attacker = tree_model.pet_index[attacker_id]
        else:
            attacker = tree_model.add_pet(
                    line.source_name + get_handle_from_id(line.source_id), attacker_id, attacker)

    ability_id = line.event_name
    if ability_id in tree_model.ability_index[attacker_id]:
        ability = tree_model.ability_index[attacker_id][ability_id]
    else:
        ability = tree_model.add_ability(line.event_name, ability_id, attacker, attacker_id)

    ability_target_id = line.target_id
    if ability_target_id in tree_model.target_index[attacker_id][ability_id]:
        ability_target = tree_model.target_index[attacker_id][ability_id][ability_target_id]
    else:
        ability_target = tree_model.add_target(row_constructor(
                line.target_name,
                get_handle_from_id(ability_target_id), ability_target_id), ability_target_id,
                ability, ability_id, attacker_id, parse_duration)

    return ability_target, ability_target.data


def get_incoming_target_row(
        tree_model: TreeModel, line: LogLine, player_attacked: bool,
        row_constructor: AnalysisTableRow,
        parse_duration: int) -> tuple[TreeItem, DamageTableRow | HealTableRow]:
    """
    Adds the needed parents to the tree model and returns the newly created or already existing
    data row. Also updates combat time of the actor.

    Parameters:
    - :param tree_model: model to work on
    - :param line: log line that contains the ability
    - :param player_attacked: True when player is attacked, False otherwise
    - :param row_constructor: a new object returned by this constructor will hold the row data
    - :param parse_duration: seconds between the first and last line of the combat, rounded up

    :return: reference to newly created or existing TreeItem and data row
    """
    target_id = (line.target_id,)
    target_handle = get_handle_from_id(line.target_id)
    if target_id in tree_model.actor_index:
        target = tree_model.actor_index[target_id]
    else:
        target = tree_model.add_actor(
                line.target_name, target_handle, target_id, row_constructor, player_attacked)
        target.data.combat_start = line.timestamp
    target.data.combat_end = line.timestamp

    if line.source_name:
        source_id = line.source_id
        source_name = line.source_name
    else:
        source_id = line.owner_id
        source_name = line.owner_name

    if source_id in tree_model.source_index[target_id]:
        ability_source = tree_model.source_index[target_id][source_id]
    else:
        ability_source = tree_model.add_source_actor(
                (source_name, get_handle_from_id(source_id)), source_id, target, target_id)

    source_ability_id = line.source_id + line.event_id
    if source_ability_id in tree_model.ability_index[target_id][source_id]:
        source_ability = tree_model.ability_index[target_id][source_id][source_ability_id]
    else:
        source_ability = tree_model.add_source_ability(
                row_constructor(line.event_name, '', source_ability_id), source_ability_id,
                ability_source, source_id, target_id, parse_duration)

    return source_ability, source_ability.data


def get_flags(flag_str: str) -> tuple[bool]:
    """
    Returns flags from flag field of log line.

    :return: (critical_hit, miss, flank, kill, shield_break)
    """
    critical_hit = 'Critical' in flag_str
    miss = 'Miss' in flag_str
    flank = 'Flank' in flag_str
    kill = 'Kill' in flag_str
    shield_break = None  # 'ShieldBreak' in flag_str
    no_floater = None  # 'NoFloater' in flag_str
    return (critical_hit, miss, flank, kill, shield_break, no_floater)


def calculate_damage_row_stats(row: DamageTableRow, combat_time: float) -> tuple:
    """
    Calculates combat stats from raw stats taken from the parse, for example DPS from damage and
    combat time.

    Parameters:
    - :param row: contains the raw data of a row
    - :param combat_time: combat time of the parent entity in seconds

    :return: complete row data according to TREE_HEADER (see __init__.py), plus a tuple containing
    name and handle as first element -> ((name, handle), DPS, Total Damage, ...)
    """
    successful_attacks = row.hull_attacks - row.misses
    try:
        dps = row.total_damage / combat_time
        shield_dps = row.total_shield_damage / combat_time
        hull_dps = row.total_hull_damage / combat_time
        base_dps = row.total_base_damage / combat_time
    except ZeroDivisionError:
        dps = 0.0
        shield_dps = 0.0
        hull_dps = 0.0
        base_dps = 0.0
    try:
        debuff = (row.total_damage / row.total_base_damage) - 1
    except ZeroDivisionError:
        debuff = 0.0
    try:
        crit_chance = row.crit_num / successful_attacks
        flank_rate = row.flank_num / successful_attacks
    except ZeroDivisionError:
        crit_chance = 0.0
        flank_rate = 0.0
    try:
        accuracy = successful_attacks / row.hull_attacks
    except ZeroDivisionError:
        accuracy = 0.0
    return ((row.name, row.handle), dps, row.total_damage, debuff, row.max_one_hit, crit_chance,
            accuracy, flank_rate, row.kills, row.total_attacks, row.misses, row.crit_num,
            row.flank_num, row.total_shield_damage, shield_dps, row.total_hull_damage, hull_dps,
            row.total_base_damage, base_dps, combat_time, row.hull_attacks, row.shield_attacks)


def calculate_heal_row_stats(row: HealTableRow, combat_time: float) -> tuple:
    """
    Calculates combat stats from raw stats taken from the parse, for example HPS from Heal and
    combat time.

    Parameters:
    - :param raw_row_data: contains the raw data of a row
    - :param combat_time: combat time of the parent entity in seconds

    :return: complete row data according to HEAL_TREE_HEADER (see constants.py), plus a tuple
    containing name and handle as first element -> ((name, handle), HPS, Total Heal, ...)
    """
    try:
        hps = row.total_heal / combat_time
        shield_hps = row.shield_heal / combat_time
        hull_hps = row.hull_heal / combat_time
    except ZeroDivisionError:
        hps = 0.0
        shield_hps = 0.0
        hull_hps = 0.0
    try:
        crit_chance = row.critical_heals / row.hull_heal_ticks
    except ZeroDivisionError:
        crit_chance = 0.0

    return ((row.name, row.handle), hps, row.total_heal, row.hull_heal, hull_hps, row.shield_heal,
            shield_hps, row.max_one_heal, crit_chance, row.heal_ticks, row.critical_heals,
            combat_time, row.hull_heal_ticks, row.shield_heal_ticks)


def combine_children_damage_stats(item: TreeItem):
    """
    Combines the stats of the children items intelligently. Absolute numbers are summed,
    percentages are averaged.

    Parameters:
    - :param item: item to retrieve children data from; item.data must be string containing the
    ability name
    """
    children_data = tuple(zip(*item._children))
    result_data = [None] * 22
    combat_time = result_data[19] = children_data[19][0]
    if isinstance(item.data, (str, tuple)):
        result_data[0] = item.data
    else:
        result_data[0] = (item.data.name, item.data.handle, item.data.id[0])
    for index in (2, 8, 9, 10, 11, 12, 13, 15, 17, 20, 21):
        result_data[index] = sum(children_data[index])
    result_data[4] = max(children_data[4])  # max_one_hit
    try:
        result_data[1] = result_data[2] / combat_time  # DPS
        result_data[14] = result_data[13] / combat_time  # shield_DPS
        result_data[16] = result_data[15] / combat_time  # hull_DPS
        result_data[18] = result_data[17] / combat_time  # base_DPS
    except ZeroDivisionError:
        result_data[1] = 0.0
        result_data[14] = 0.0
        result_data[16] = 0.0
        result_data[18] = 0.0
    try:
        result_data[3] = (result_data[2] / result_data[17]) - 1  # debuff
    except ZeroDivisionError:
        result_data[3] = 0.0
    successful_attacks = result_data[20] - result_data[10]
    try:
        result_data[5] = result_data[11] / successful_attacks  # crit_chance
        result_data[6] = successful_attacks / result_data[20]  # accuracy
        result_data[7] = result_data[12] / successful_attacks  # flank_rate
    except ZeroDivisionError:
        result_data[5] = 0.0
        result_data[6] = 0.0
        result_data[7] = 0.0
    item.data = tuple(result_data)


def combine_children_heal_stats(item: TreeItem):
    """
    Combines the stats of the children items intelligently. Absolute numbers are summed,
    percentages are averaged.

    Parameters:
    - :param item: item to retrieve children data from; item.data must be string containing the
    ability name
    """
    children_data = tuple(zip(*item._children))
    result_data = [None] * 14
    combat_time = result_data[11] = children_data[11][0]
    if isinstance(item.data, (str, tuple)):
        result_data[0] = item.data
    else:
        result_data[0] = (item.data.name, item.data.handle, item.data.id[0])
    for index in (2, 3, 5, 9, 10, 12, 13):
        result_data[index] = sum(children_data[index])
    result_data[7] = max(children_data[7])  # max_one_heal
    try:
        result_data[1] = result_data[2] / combat_time  # HPS
        result_data[4] = result_data[3] / combat_time  # hull_HPS
        result_data[6] = result_data[5] / combat_time  # shield_HPS
    except ZeroDivisionError:
        result_data[1] = 0.0
        result_data[4] = 0.0
        result_data[6] = 0.0
    try:
        result_data[8] = result_data[10] / result_data[12]  # critical_heals
    except ZeroDivisionError:
        result_data[8] = 0.0
    item.data = tuple(result_data)


def merge_single_lines(tree_model: TreeModel):
    """
    Eliminates one level of depth if needed by merging lines that only have a single child. For
    example:

    v Quantum Mines
        v Quantum Mine 34
            > Mine Explosion
        v Quantum Mine 25
            > Mine Explosion

    becomes:\n
    v Quantum Mines - Mine Explosion\n
       > Quantum Mine 34\n
       > Quantum Mine 25


    Parameters:
    - :param tree_model: tree model to analyze and improve
    """
    for player in tree_model._player._children:
        new_pet_groups = dict()

        for ability_or_petgroup in player._children:
            if (len(ability_or_petgroup._children) < 1
                    or ability_or_petgroup._children[0].child_count == 0):
                continue

            for pet in reversed(ability_or_petgroup._children):
                if pet.child_count == 1:
                    ability_name = pet._children[0].data
                    if ability_or_petgroup.data == ability_name:
                        new_pet_group_name = ability_name
                    else:
                        new_pet_group_name = f'{ability_or_petgroup.data} – {ability_name}'
                    if new_pet_group_name in new_pet_groups:
                        new_pet_group = new_pet_groups[new_pet_group_name]
                    else:
                        new_pet_group = TreeItem(new_pet_group_name, player)
                        new_pet_groups[new_pet_group_name] = new_pet_group
                    new_pet_group.append_child(pet)
                    ability_or_petgroup._children.remove(pet)
                    pet.parent = new_pet_group
                    pet._children = pet._children[0]._children

                    for target in pet._children:
                        target.parent = pet

            if ability_or_petgroup.child_count == 0:
                ability_or_petgroup.parent._children.remove(ability_or_petgroup)

        for new_pet_group in new_pet_groups.values():
            player.append_child(new_pet_group)


def complete_damage_sub_tree(item: TreeItem, combat_time):
    """
    Recursive function merging data from the bottom up.

    Parameters:
    - :param item: item to complete
    """
    graph_data = list()
    for child in item._children:
        if child.child_count == 0:
            child.data = calculate_damage_row_stats(child.data, combat_time)
        else:
            complete_damage_sub_tree(child, combat_time)
        graph_data.append(child.graph_data)
    combine_children_damage_stats(item)
    item.graph_data = numpy.sum(graph_data, axis=0, dtype=numpy.float64)


def complete_heal_sub_tree(item: TreeItem, combat_time):
    """
    Recursive function merging data from the bottom up.

    Parameters:
    - :param item: item to complete
    """
    graph_data = list()
    for child in item._children:
        if child.child_count == 0:
            child.data = calculate_heal_row_stats(child.data, combat_time)
        else:
            complete_heal_sub_tree(child, combat_time)
        graph_data.append(child.graph_data)
    combine_children_heal_stats(item)
    item.graph_data = numpy.sum(graph_data, axis=0, dtype=numpy.float64)


def complete_damage_tree(
        tree_model: TreeModel, combat_durations: dict, total_combat_duration: float):
    """
    Merges the data from the bottom up to fill all lines.

    Parameters:
    - :param tree_model: tree model to be completed
    - :param combat_durations: combat durations for all actors
    """
    for actor in bundle(tree_model._player._children, tree_model._npc._children):
        current_combat_time = combat_durations.get(actor.data.id[0], total_combat_duration)
        actor.data.combat_time = current_combat_time
        complete_damage_sub_tree(actor, current_combat_time)


def complete_heal_tree(tree_model: TreeModel, combat_durations: dict, total_combat_duration: float):
    """
    Merges the data from the bottom up to fill all lines.

    Parameters:
    - :param tree_model: tree model to be completed
    - :param combat_durations: combat durations for all actors
    """
    for actor in bundle(tree_model._player._children, tree_model._npc._children):
        current_combat_time = combat_durations.get(actor.data.id[0], total_combat_duration)
        actor.data.combat_time = current_combat_time
        complete_heal_sub_tree(actor, current_combat_time)



from .constants import HEAL_TREE_HEADER, LIVE_TABLE_HEADER, TABLE_HEADER, TREE_HEADER
from .datamodels import TreeItem
from .iofunc import compose_logfile, extract_bytes, repair_logfile
from .liveparser import LiveParser
from .main import OSCR

__all__ = (
        'compose_logfile', 'extract_bytes', 'HEAL_TREE_HEADER', 'LIVE_TABLE_HEADER', 'LiveParser',
        'OSCR', 'repair_logfile', 'TABLE_HEADER', 'TREE_HEADER', 'TreeItem')



from datetime import datetime, timedelta
from typing import Generator, Iterable


def to_datetime(date_time: str) -> datetime:
    """
    returns datetime object from combatlog string containing date and time
    """
    date_time_list = date_time.split(':')
    date_time_list += date_time_list.pop().split('.')
    date_time_list = list(map(int, date_time_list))
    date_time_list[0] += 2000
    date_time_list[6] *= 100000
    return datetime(*date_time_list)


def to_microseconds(timedelta: timedelta):
    return ((timedelta.days * (24 * 3600) + timedelta.seconds) * 1000000
            + timedelta.microseconds)


def datetime_to_display(date_time: datetime) -> str:
    """
    Converts datetime object to formatted string.
    """
    return (f'{date_time.year}-{date_time.month:02d}-{date_time.day:02d} {date_time.hour:02d}:'
            f'{date_time.minute:02d}:{date_time.second:02d}')


def get_handle_from_id(id_str: str) -> str:
    """
    returns player handle from id string
    """
    if id_str.startswith('P'):
        _, _, handle_part = id_str.rpartition('@')
        return '@' + handle_part[:-1]
    elif id_str.startswith('C'):
        handle_part, _ = id_str.split(' ', maxsplit=1)
        # the space enables consistent concatenation behaviour between players and NPCs
        return f' {handle_part[2:]}'
    else:
        return ''


def get_player_handle(id_str: str) -> str:
    """
    returns player handle from id string
    """
    _, _, handle_part = id_str.rpartition('@')
    return '@' + handle_part[:-1]


def get_entity_name(entity_id: str) -> str:
    """
    Returns the entity name matching an entity id.
    """
    split = entity_id.split(" ")
    if len(split) != 2:
        return None
    return split[1].replace("]", "")


def bundle[_T](*iterables: Iterable[_T]) -> Generator[_T, None, None]:
    """
    Generator yielding the items of the given iterables in the order they were provided.

    Parameters:
    - :param iterables: iterables to be bundled
    """
    for inner_iterable in iterables:
        for element in inner_iterable:
            yield element



from collections.abc import Callable
from datetime import timedelta, datetime
from multiprocessing import Event, Process, Queue
from multiprocessing.pool import Pool
import os
from queue import Empty as EmptyException

from .combat import Combat
from .constants import BANNED_ABILITIES
from .datamodels import LogLine, TreeItem
from .detection import Detection
from .iofunc import extract_bytes, reset_temp_folder
from .oscr_read_file_backwards import ReadFileBackwards
from .parser import analyze_combat
from .utilities import datetime_to_display, get_entity_name, to_datetime


def _f(*args, **kwargs):
    pass


def raise_error(error: BaseException):
    raise error


class OSCR:

    __version__ = '10.3.2'

    def __init__(self, log_path: str = '', settings: dict = None):
        self.log_path = log_path
        self.combats: list[Combat] = list()
        self.bytes_consumed: int = 0  # -1 would mean entire log has been consumed
        self.combat_analyzed_callback: Callable[[Combat], None] = _f
        self.error_callback: Callable[[BaseException], None] = raise_error
        self._settings = {
            "combats_to_parse": 10,
            "seconds_between_combats": 100,
            "combat_min_lines": 20,
            "excluded_event_ids": ["Autodesc.Combatevent.Falling"],
            "graph_resolution": 0.2,
            "templog_folder_path": f"{os.path.dirname(os.path.abspath(__file__))}/~temp_log_files",
        }
        self._pool = None
        self._queue = None

        if settings is not None:
            self._settings.update(settings)
        reset_temp_folder(self._settings['templog_folder_path'])

    def __del__(self):
        if isinstance(self._pool, Pool):
            self._pool.terminate()

    @property
    def analyzed_combats(self) -> list[str]:
        """
        Contains tuple with available combats.
        """
        res = list()
        for c in self.combats:
            if c.difficulty:
                res.append(
                    f"{c.map} ({c.difficulty} Difficulty) at {datetime_to_display(c.start_time)}"
                )
            else:
                res.append(f"{c.map} {datetime_to_display(c.start_time)}")
        return res

    def reset_parser(self):
        """
        Resets the parser to default state. Removes stored combats, logfile data and log path.
        """
        self.log_path = ''
        self.combats = list()
        self.bytes_consumed = 0

    @staticmethod
    def _analyze_log_file(
            log_path: str, total_combats: int, first_combat_id: int, offset: int, settings: dict,
            combat_handler: Callable[[Combat], None] = _f,
            error_handler: Callable[[BaseException], None] = raise_error):
        """
        (Internal Function) Reads a logfile, isolates combats and calls `combat_handler` for each
        combat as soon as it has been found.

        Parameters:
        - :param log_path: log path to be analyzed; overwrites `self.log_path`
        - :param total_combats: stops isolating combats when combat id reaches `total_combats`
        - :param first_combat_id: id that the first found combat gets; id is incremented per combat
        - :param offset: offset in bytes from the end of the logfile
        - :param settings: contains settings for parser; uses keys "seconds_between_combats" and
        "graph_resolution"
        - :param combat_handler: Called once for each analyzed combat as soon as the combats
        analyzation is complete
        - :param error_handler: Called when an error occurs during logfile analyzation

        :return: -1 if entire file has been consumed; otherwise next byte to analyze counted from
        the end of the file
        """
        combat_delta = timedelta(seconds=settings['seconds_between_combats'])
        combat_id = first_combat_id
        current_combat = Combat(settings['graph_resolution'], combat_id, log_path)
        log_consumed = True
        try:
            with ReadFileBackwards(log_path, offset) as backwards_file:
                if len(backwards_file.top) <= 2:
                    last_log_time = datetime.now() + timedelta(days=1)
                else:
                    last_log_time = to_datetime(backwards_file.top.split('::')[0])
                current_combat.end_time = last_log_time
                current_combat.file_pos[1] = backwards_file.filesize - offset
                for line in backwards_file:
                    if len(line) <= 2:
                        continue
                    time_data, attack_data = line.split('::')
                    splitted_line = attack_data.split(',')
                    if splitted_line[6] in BANNED_ABILITIES:
                        continue
                    log_time = to_datetime(time_data)
                    if last_log_time - log_time > combat_delta:
                        current_file_position = backwards_file.filesize - (
                                backwards_file.get_bytes_read(True) + offset)
                        if len(current_combat.log_data) >= settings['combat_min_lines']:
                            current_combat.start_time = last_log_time
                            current_combat.file_pos[0] = current_file_position
                            combat_handler(current_combat)
                            combat_id += 1
                        if combat_id >= total_combats:
                            log_consumed = False
                            new_offset = backwards_file.get_bytes_read(True) + offset
                            break
                        current_combat = Combat(settings['graph_resolution'], combat_id, log_path)
                        current_combat.end_time = log_time
                        current_combat.file_pos[1] = current_file_position
                    current_line = LogLine(
                        log_time,
                        *splitted_line[:10],
                        float(splitted_line[10]),
                        float(splitted_line[11]),
                    )
                    last_log_time = log_time
                    current_combat.log_data.appendleft(current_line)
            if log_consumed:
                if len(current_combat.log_data) >= settings['combat_min_lines']:
                    current_combat.start_time = log_time
                    current_combat.file_pos[0] = 0
                    combat_handler(current_combat)
                new_offset = -1
        except BaseException as e:
            if 'line' in locals():
                e.args = (*e.args, line)
            else:
                e.args = (*e.args, 'Error before loop!')
            error_handler(e)
            return 0
        return new_offset

    def analyze_log_file(
            self, log_path: str = '', max_combats: int = -1, offset: int = -1,
            result_handler: Callable[[Combat], None] = _f):
        """
        Analyzes log file in `self.log_file` and appends analyzed combats to `self.combats`.
        (Can cause duplicate combats to appear, use `self.reset_parser` if analyzing new log file.)

        Parameters:
        - :param log_path: log path to be analyzed; overwrites `self.log_path`
        - :param max_combats: maximum number of combats to analyze
        - :param offset: offset in bytes from the end of the logfile
        - :param result_handler: Called once for each analyzed combat as soon as the combats
        analyzation is complete
        """

        if log_path != '':
            self.log_path = log_path
        elif self.log_path == '':
            raise AttributeError(
                '"self.log_path" or parameter "log_path" must contain a path to a log file.'
            )
        if self.bytes_consumed < 0:
            return
        next_combat_id = len(self.combats)
        if max_combats < 0:
            max_combats = self._settings['combats_to_parse']
        total_combats = len(self.combats) + max_combats
        self.combats.extend([None] * max_combats)
        if offset < 0:
            offset = self.bytes_consumed
        if result_handler is not _f:
            self.combat_analyzed_callback = result_handler
        self.bytes_consumed = OSCR._analyze_log_file(
                self.log_path, total_combats, next_combat_id, offset, self._settings,
                self.analyze_new_combat, self.error_callback)
        self.combats = [combat for combat in self.combats if combat is not None]

    def analyze_log_file_mp(
            self, log_path: str = '', max_combats: int = -1, offset: int = -1,
            result_handler: Callable[[Combat], None] = _f):
        """
        Analyzes log file in `self.log_file` and appends analyzed combats to `self.combats`.
        (Can cause duplicate combats to appear, use `self.reset_parser` if analyzing new log file.)
        Blocks until given number of combats have been isolated.

        Parameters:
        - :param log_path: log path to be analyzed; overwrites `self.log_path`
        - :param max_combats: maximum number of combats to analyze
        - :param offset: offset in bytes from the end of the logfile
        - :param result_handler: Called once for each analyzed combat as soon as the combats
        analyzation is complete
        """
        if log_path != '':
            self.log_path = log_path
        elif self.log_path == '':
            raise AttributeError(
                '"self.log_path" or parameter "log_path" must contain a path to a log file.'
            )
        if self.bytes_consumed < 0:
            return
        next_combat_id = len(self.combats)
        if max_combats < 0:
            max_combats = self._settings['combats_to_parse']
        total_combats = len(self.combats) + max_combats
        self.combats.extend([None] * max_combats)
        if offset < 0:
            offset = self.bytes_consumed
        if result_handler is not _f:
            self.combat_analyzed_callback = result_handler
        self._pool = Pool(4)
        self._queue = Queue()
        args = (self._queue, self.log_path, total_combats, next_combat_id, offset, self._settings)
        logfile_process = Process(target=OSCR._analyze_file_helper, args=args)
        logfile_process.start()
        combats_isolated_num = 0
        while True:
            try:
                data = self._queue.get(timeout=15)
            except EmptyException:
                break
            if isinstance(data, int):
                self.bytes_consumed = data
                for _ in range(max_combats - combats_isolated_num):
                    self.combats.pop()
                break
            if isinstance(data, BaseException):
                self._pool.terminate()
                self.error_callback(data)
                break
            else:
                combats_isolated_num += 1
                self._pool.apply_async(
                        analyze_combat, args=(data,), callback=self.handle_analyzed_result,
                        error_callback=self.error_callback)
        self._pool.close()

    @staticmethod
    def _analyze_file_helper(queue, log_path, total_combats, first_combat_id, offset, settings):
        """
        Helper method to put return value of function into queue to be sent to main process. Wraps
        `_analyze_log_file`.
        """
        queue.put(OSCR._analyze_log_file(
                log_path, total_combats, first_combat_id, offset, settings,
                lambda combat: queue.put(combat), lambda error: queue.put(error)))

    def analyze_new_combat(self, combat: Combat):
        """
        Analyzes isolated combat, puts it into `self.combats` and calls the combat analyzed callback
        """
        analyze_combat(combat)
        self.combats[combat.id] = combat
        self.combat_analyzed_callback(combat)

    def handle_analyzed_result(self, result_combat: Combat):
        """
        puts analyzed combat into `self.combats` and calls the combat analyzed callback
        """
        self.combats[result_combat.id] = result_combat
        self.combat_analyzed_callback(result_combat)

    def isolate_combats(self, path: str, max_combats: int = -1) -> list[tuple]:
        """
        Returns list of combats in logfile at given `path`.

        Parameters:
        - :param path: path to the logfile
        - :param max_combats: maximum number of combats to isolate

        :return: tuple(number of combat in file, map, date, time, difficulty, byte_start, byte_end)
        """
        combat_delta = timedelta(seconds=self._settings['seconds_between_combats'])
        combat_id = 0
        current_map_and_difficulty = ['Combat', '']
        map_identifiers = set(Detection.MAP_IDENTIFIERS_EXISTENCE.keys())
        combats = list()
        current_end_bytes = -1
        current_line_num = 0
        try:
            with ReadFileBackwards(path) as backwards_file:
                if len(backwards_file.top) <= 2:
                    llt = datetime.now() + timedelta(days=1)
                else:
                    llt = to_datetime(backwards_file.top.split('::')[0])
                current_end_bytes = backwards_file.filesize
                for line in backwards_file:
                    if len(line) <= 2:
                        continue
                    time_data, attack_data = line.split('::')
                    splitted_line = attack_data.split(',')
                    log_time = to_datetime(time_data)
                    if llt - log_time > combat_delta:
                        current_file_position = backwards_file.filesize - (
                                backwards_file.get_bytes_read(True))
                        if current_line_num >= self._settings['combat_min_lines']:
                            combats.append((
                                combat_id,
                                current_map_and_difficulty[0],
                                f'{llt.year}-{llt.month:02d}-{llt.day:02d}',
                                f'{llt.hour:02d}:{llt.minute:02d}:{llt.second:02d}',
                                current_map_and_difficulty[1],
                                current_file_position,
                                current_end_bytes))
                            combat_id += 1
                        if combat_id >= max_combats > 0:
                            return combats
                        current_map_and_difficulty = ['Combat', '']
                        current_end_bytes = current_file_position
                        current_line_num = 0
                    if get_entity_name(splitted_line[5]) in map_identifiers:
                        m = Detection.MAP_IDENTIFIERS_EXISTENCE[get_entity_name(splitted_line[5])]
                        current_map_and_difficulty[0] = m['map']
                        if m['difficulty'] != 'Any':
                            current_map_and_difficulty[1] = m['difficulty']
                    current_line_num += 1
                    llt = log_time
            if current_line_num >= self._settings['combat_min_lines']:
                combats.append((
                    combat_id,
                    current_map_and_difficulty[0],
                    f'{llt.year}-{llt.month:02d}-{llt.day:02d}',
                    f'{llt.hour:02d}:{llt.minute:02d}:{llt.second:02d}',
                    current_map_and_difficulty[1],
                    backwards_file.filesize - backwards_file.get_bytes_read(),
                    current_end_bytes))
        except BaseException as e:
            if 'line' in locals():
                e.args = (*e.args, line)
            else:
                e.args = (*e.args, 'Error before loop!')
            self.error_callback(e)
            return list()
        return combats

    def export_combat(self, combat_num: int, path: str):
        """
        Exports combat to new logfile

        Parameters:
        - :param combat_num: index of the combat in self.combats
        - :param path: path to export the log to, will overwrite existing files
        """
        try:
            combat = self.combats[combat_num]
        except IndexError:
            raise AttributeError(
                f"Combat #{combat_num} you are trying to save has not been isolated yet."
                f"Number of isolated combats: {len(self.combats)} -- Use "
                "OSCR.analyze_log_file() with appropriate arguments first.")
        extract_bytes(combat.log_file, path, combat.file_pos[0], combat.file_pos[1])



""" OSCR CLI """

from argparse import ArgumentParser
import cProfile
from datetime import datetime
import os
from pathlib import Path
import pstats

from . import OSCR
from .combat import Combat


OVERVIEW_HEADER = [
    'Player', 'DPS', 'Combat Time', 'Combat Time Share', 'Total Damage', 'Debuff',
    'Attacks-in Share', 'Taken Damage Share', 'Damage Share', 'Deaths']

HELP = """OSCR CLI Usage:
• help, h
    Shows this help message.
• open <file>, o <file>
    Selects file to analyze.
• overview <combat>, ov <combat>
    Analyzes specified combat identified by combat ID and shows
    overview table. Not specifying a combat will analyze the most
    recent combat.
• combats <amount>, c <amount>
    Shows specified amount of combats in selected logfile. Shows
    all combats in selected log file if no argument is given.
• quit, q
    Exits the interactive mode and ends the program."""


def convert_datetime(dt: datetime) -> tuple[str, str]:
    """
    Extracts date and time from datetime object and returns them separately in formatted form.
    """
    return (f'{dt.year}-{dt.month:02d}-{dt.day:02d}',
            f'{dt.hour:02d}:{dt.minute:02d}:{dt.second:02d}')


def format_table(table: list[list[str]], header: list[str], column_alignment: list[str]) -> str:
    """
    Puts table consisting of nested lists into str format for output on the terminal. Truncates
    width of table if it is wider than the current terminal width.

    Parameters:
    - :param table: table to convert; outer list contains rows, each inner list contains the
    columns for that row; each row must be of the same length and match the length of `header`
    - :param header: labels for the column heading
    - :param column_alignment: `r` or `l` to align column content except header right or left,
    respectively; any other value will center-align the content; must match the length of `header`
    """
    column_widths = [len(e) for e in header]
    output = list()
    for row in table:
        for i, item in enumerate(row):
            column_widths[i] = max(column_widths[i], len(str(item)))  # TODO unify type beforehand
    border = '┌'
    for width in column_widths:
        border += '─' * width + '┬'
    border = border[:-1] + '┐'
    output.append(border)
    formatted_row = ''
    for item, width in zip(header, column_widths):
        formatted_row += f'│{item:^{width}}'
    output.append(formatted_row + '│')
    border = '├'
    for width in column_widths:
        border += '─' * width + '┼'
    border = border[:-1] + '┤'
    output.append(border)
    for row in table:
        formatted_row = ''
        for column, item in enumerate(row):
            match column_alignment[column]:
                case 'l':
                    formatted_row += f'│{item:<{column_widths[column]}}'
                case 'r':
                    formatted_row += f'│{item:>{column_widths[column]}}'
                case _:
                    formatted_row += f'│{item:^{column_widths[column]}}'
        output.append(formatted_row + '│')
    border = '└'
    for width in column_widths:
        border += '─' * width + '┴'
    border = border[:-1] + '┘'
    output.append(border)
    width = max(69, os.get_terminal_size().columns - 1)
    if width < len(output[0]):
        return '\n'.join(map(lambda row: row[:width], output))
    else:
        return '\n'.join(output)


def open_logfile(
        parser: OSCR, isolated_combats: list[list], logfile_path: str,
        max_combats: int = 5) -> bool:
    """
    """
    path = Path(logfile_path)
    if not path.exists() or not path.is_file():
        print('The specified logfile does not exist. Please choose a different file.')
        return False
    parser.reset_parser()
    parser.log_path = logfile_path
    raw_combats = parser.isolate_combats(logfile_path, max_combats)
    if len(raw_combats) < 1:
        print('The specified logfile does not contain any combats.')
        return False
    isolated_combats.clear()
    for combat in map(list, raw_combats):
        combat[0] += 1
        isolated_combats.append(combat[:5])
    return True


def get_overview_data(combat: Combat, sort_column: int = 1) -> list[list]:
    """
    """
    player_data = [p for p in combat.players.values()]
    if sort_column != 0:
        sort_column += 1
    player_data.sort(key=lambda row: row[sort_column], reverse=True)
    data = list()
    for player in player_data:
        data.append([
            player.name + player.handle,
            f'{player.DPS:,.2f}',
            f'{player.combat_time:.1f}s',
            f'{player.combat_time_share * 100:.2f}%',
            f'{player.total_damage:,.2f}',
            f'{player.debuff * 100:.2f}%',
            f'{player.attacks_in_share * 100:.2f}%',
            f'{player.taken_damage_share * 100:.2f}%',
            f'{player.damage_share * 100:.2f}%',
            f'{player.deaths}'
        ])
    return data


def interpret_combat_argument(args: list[str], available_count: int) -> int:
    """
    """
    count = 0
    if len(args) > 0:
        try:
            count = int(args[0])
        except ValueError:
            print(
                'Invalid input. If supplied, the argument must be an integer. '
                'Using default value...')
    if count > 0:
        return count
    elif available_count > 0:
        return available_count
    else:
        return count


def get_combats(parser: OSCR, isolated_combats: list[list], max_combats: int = 5) -> bool:
    """
    """
    if parser.log_path == '':
        print('No logfile specified. Please open a logfile first.')
        return False
    else:
        raw_combats = parser.isolate_combats(parser.log_path, max_combats)
        if len(raw_combats) < 1:
            print('The specified logfile does not contain any combats.')
            return False
        for combat in map(list, raw_combats[len(isolated_combats):]):
            combat[0] += 1
            isolated_combats.append(combat[:5])
    return True


def produce_overview_data(parser: OSCR, isolated_combats: list[list], combat_id: int) -> list[list]:
    """
    """
    if combat_id >= len(parser.combats):
        parser.analyze_log_file(max_combats=combat_id - len(parser.combats) + 1)
        for id, combat in enumerate(parser.combats):
            combat_summary = [
                combat.id + 1,
                combat.map,
                *convert_datetime(combat.start_time),
                combat.difficulty if combat.difficulty is not None else ''
            ]
            if id < len(isolated_combats):
                isolated_combats[id] = combat_summary
            else:
                isolated_combats.append(combat_summary)
        if combat_id >= len(parser.combats):
            print(
                'The specified combat does not exist in the given logfile. '
                'Use "combats" / "--combats" to show available combats.')
            return []
    combat = parser.combats[combat_id]
    return get_overview_data(combat)


def print_combat_announcer(combat: Combat):
    """
    """
    formatted_date, formatted_time = convert_datetime(combat.start_time)
    if combat.difficulty is None:
        difficulty = ''
    else:
        difficulty = f' ({combat.difficulty})'
    print(f'[{combat.id + 1}] -> {combat.map}{difficulty} '
          f'{formatted_date} {formatted_time}')


def interactive_cli(log_path: str | None = None):
    """
    Executes interactive cli for OSCR.
    """
    parser = OSCR()
    # TODO this needs to be integrated into OSCR, but that requires a rewrite
    isolated_combats = list()
    print(
        '      >>  Open Source Combatlog Reader  <<\n\nWelcome to the interactive CLI of OSCR '
        f'{OSCR.__version__}!\nType "help" for more information, "quit" to quit.')
    if log_path is not None and open_logfile(parser, isolated_combats, log_path):
        print(format_table(
            isolated_combats, ['ID', 'Map', 'Date', 'Time', 'Difficulty'],
            ['r', 'l', 'l', 'l', 'l']))
    while True:
        try:
            message = input('(OSCR) ')
        except EOFError:
            break
        action, *arguments = message.split(' ')
        if ''.join(arguments) == '':
            arguments = []
        match action:
            case 'q' | 'quit':
                break
            case 'h' | 'help':
                print(HELP)
            case 'open' | 'o':
                if open_logfile(parser, isolated_combats, ''.join(arguments)):
                    print(format_table(
                        isolated_combats, ['ID', 'Map', 'Date', 'Time', 'Difficulty'],
                        ['r', 'l', 'l', 'l', 'l']))
            case 'combats' | 'c':
                combats_to_show = interpret_combat_argument(arguments, len(isolated_combats))
                if len(isolated_combats) >= combats_to_show:
                    print(format_table(
                        isolated_combats[:combats_to_show],
                        ['ID', 'Map', 'Date', 'Time', 'Difficulty'], ['r', 'l', 'l', 'l', 'l']))
                else:
                    if get_combats(parser, isolated_combats, combats_to_show):
                        print(format_table(
                            isolated_combats, ['ID', 'Map', 'Date', 'Time', 'Difficulty'],
                            ['r', 'l', 'l', 'l', 'l']))
            case 'overview' | 'ov':
                combat_to_show = interpret_combat_argument(arguments, 1) - 1
                data = produce_overview_data(parser, isolated_combats, combat_to_show)
                print_combat_announcer(parser.combats[combat_to_show])
                print(format_table(
                    data, OVERVIEW_HEADER,
                    ['l', 'r', 'r', 'r', 'r', 'r', 'r', 'r', 'r', 'r']))
            case _:
                print('Invalid command. Enter "help" to see available commands.')
                continue


def main():
    """Main"""
    argparser = ArgumentParser(prog='OSCR', description='CLI for the Open Source Combatlog Reader')
    argparser.add_argument(
        '--open', '-o', type=str, required=False, metavar='PATH', help='Selects file to analyze.')
    argparser.add_argument(
        '--combats', '-c', type=int, required=False, metavar='ID', nargs='?', const=5,
        help='Lists given amount of combats from the selected logfile.')
    argparser.add_argument(
        '--overview', '-ov', type=int, required=False, metavar='ID', nargs='?', const=1,
        help='Shows overview table for given combat.')
    args, _ = argparser.parse_known_args()
    if args.open is None:
        if args.combats is None and args.overview is None:
            interactive_cli()
        else:
            print('Could not perform action: --open [PATH] must be specified')
    elif args.combats is None and args.overview is None:
        interactive_cli(args.open)
    else:
        parser = OSCR()
        isolated_combats = list()
        combats_to_show = 5
        if args.combats is not None and args.combats > 0:
            combats_to_show = args.combats
        if not open_logfile(parser, isolated_combats, args.open, combats_to_show):
            return
        if args.overview is None:
            print(format_table(
                isolated_combats, ['ID', 'Map', 'Date', 'Time', 'Difficulty'],
                ['r', 'l', 'l', 'l', 'l']))
        else:
            combat_to_show: int = args.overview - 1
            if combat_to_show < 0:
                print('The combat ID must be at least 1.')
                return
            data = produce_overview_data(parser, isolated_combats, combat_to_show)
            if args.combats is not None:
                print(format_table(
                    isolated_combats[:combats_to_show],
                    ['ID', 'Map', 'Date', 'Time', 'Difficulty'], ['r', 'l', 'l', 'l', 'l']))
            print_combat_announcer(parser.combats[combat_to_show])
            print(format_table(
                data, OVERVIEW_HEADER,
                ['l', 'r', 'r', 'r', 'r', 'r', 'r', 'r', 'r', 'r']))


if __name__ == "__main__":
    if os.environ.get("CPROFILE"):
        pr = cProfile.Profile()
        pr.enable()

    main()

    if os.environ.get("CPROFILE"):
        pr.disable()
        ps = pstats.Stats(pr).sort_stats("cumtime")
        print("================================")
        ps.print_stats()



"""This file implements the Combat class"""

from collections import deque
from datetime import datetime

import numpy

from .constants import HEAL_TREE_HEADER, TREE_HEADER
from .datamodels import CritterMeta, DetectionInfo, LogLine, OverviewTableRow, TreeItem, TreeModel
from .detection import Detection
from .export import analysis_table_export
from .utilities import datetime_to_display, get_entity_name


def check_difficulty_deaths(
            difficulty_data: dict, combat_meta: dict[str, CritterMeta]) -> DetectionInfo:
    """
    Check deaths against combat metadata

    Parameters:
    - :param difficulty_data: difficulty-based dictionary from `MAP_DIFFICULTY_ENTITY_DEATH_COUNTS`\
    containing required death counts of specific entities
    - :param combat_meta: Combat metadata generated by `Combat.detect_map`
    """
    for entity_name, required_death_count in difficulty_data.items():
        entity_metadata = combat_meta.get(entity_name)
        if entity_metadata is None:
            # Map is missing some NPC data - it's invalid.
            return DetectionInfo(False, 'difficulty', (entity_name,), step='existence')
        if required_death_count > 0:
            valid = required_death_count == entity_metadata.deaths
        else:
            valid = entity_metadata.deaths != 0
        if not valid:
            return DetectionInfo(
                    False, 'difficulty', (entity_name,), required_death_count,
                    entity_metadata.deaths, 'deaths')
    return DetectionInfo(True, 'difficulty', tuple(difficulty_data.keys()), step='deaths')


def check_difficulty_damage(
            difficulty_data: dict, combat_meta: dict[str, CritterMeta]) -> DetectionInfo:
    """
    Check hull damage taken against combat metadata

    Parameters:
    - :param difficulty_data: difficulty-based dicitionary from `MAP_DIFFICULTY_ENTITY_HULL_COUNTS`\
    containing required hull capacity of specific entities
    - :param combat_meta: Combat metadata generated by `Combat.detect_map`
    """
    # only look at the lower variance.
    var = 0.20

    for entity_name, hull_value in difficulty_data.items():
        entity_metadata = combat_meta.get(entity_name)
        if entity_metadata is None:
            # Map is missing some NPC data - it's invalid.
            return DetectionInfo(False, 'difficulty', step='damage')
        med = numpy.percentile(entity_metadata.hull_values, 50)
        low = hull_value * (1 - var)
        valid = low < med
        if not valid:
            return DetectionInfo(False, 'difficulty', (entity_name,), low, med, 'damage')
    return DetectionInfo(True, 'difficulty', tuple(difficulty_data.keys()), step='damage')


class Combat:
    """
    Contains a single combat including raw log lines, map and combat information and shallow parse
    results.
    """

    def __init__(self, graph_resolution: float = 0.2, id: int = -1, log_file: str = ''):
        """
        Parameters:
        - :param graph_resolution: duration of one graph interval
        - :param id: id of the combat; usually counted from the end of the logfile
        - :param log_file: path to the log file this combat was parsed from
        """
        self.log_data: deque[LogLine] = deque()
        self.id: int = id
        self.map = None
        self.difficulty = None
        self.start_time: datetime = None
        self.end_time: datetime = None
        self.file_pos: list[int, int] = [None, None]
        self.log_file: str = log_file
        self.meta: dict = {
            'log_duration': None,
            'player_duration': None,
            'detection_info': None  # iterable of DetectionInfo
        }
        self.players: dict[str, OverviewTableRow] = dict()
        self.critters: dict[str, CritterMeta] = dict()
        # self.critter_meta: dict = dict()
        self.graph_resolution = graph_resolution
        self.overview_graphs: dict = dict()
        self.damage_out: TreeModel = None
        self.damage_in: TreeModel = None
        self.heals_out: TreeModel = None
        self.heals_in: TreeModel = None

    @property
    def root_items(self):
        """Root items of analysis tree data: Damage Out, Damage In, Heals Out, Heals In"""
        return (
            self.damage_out._root,
            self.damage_in._root,
            self.heals_out._root,
            self.heals_in._root
        )

    @property
    def description(self):
        if self.difficulty is None:
            return f'{self.map} {datetime_to_display(self.start_time)}'
        return (
            f'{self.map} ({self.difficulty} Difficulty) at '
            + datetime_to_display(self.start_time)
        )

    @property
    def duration(self):
        return self.end_time - self.start_time

    @property
    def date_time(self):
        """Returns the end time - for compatibility with previous versions"""
        return self.end_time

    @property
    def player_dict(self):
        """Returns the list of players - for compatibility with previous versions"""
        return self.players

    def create_overview_graphs(self, player: OverviewTableRow, combat_interval: tuple[int, int]):
        """
        creates overview graphs from damage graph

        Parameters:
        - :param player: player to create graphs for
        - :param combat_interval: first and last graph point of the respective player (a graph
        point is an interval of length `graph_resolution` counted from the beginning of the log)
        """
        graph = self.overview_graphs[player.handle]
        player.DMG_graph_data = dmg_graph = graph[combat_interval[0]:combat_interval[1] + 1]
        first_graph_time = self.graph_resolution * (combat_interval[0] + 1)
        last_graph_time = self.graph_resolution * (combat_interval[1] + 1)
        player.graph_time = numpy.linspace(first_graph_time, last_graph_time, len(dmg_graph))
        combat_time_array = player.graph_time - self.graph_resolution * combat_interval[0]
        player.DPS_graph_data = dmg_graph.cumsum() / combat_time_array

    def create_overview(self, overview_graph_intervals: dict[str, tuple[int, int]]):
        """
        Creates overview table from analysis data and overview graphs from self.overiew_graphs and
        creates players with that data in self.players

        Parameters:
        - :overview_graph_intervals: maps player handles to their active combat start and end times
        measured in graph points (1 / graph_resolution points per second of the log)
        """
        combat_duration = self.meta['player_duration']
        if combat_duration == 0:
            return
        total_damage_out = 0
        total_attacks_in = 0
        total_damage_in = 0
        total_heals = 0
        # build_detection_abilities = tuple(Detection.BUILD_DETECTION_ABILITIES.keys())

        for player_item in self.damage_out._player._children:
            dmg_out_data = player_item.data
            player_combat_duration = dmg_out_data[19]
            if player_combat_duration <= 0:
                continue
            player_name_handle_id = dmg_out_data[0]
            player = OverviewTableRow(player_name_handle_id[0], player_name_handle_id[1])
            player.DPS = dmg_out_data[1]
            player.combat_time = player_combat_duration
            player.combat_time_share = player_combat_duration / combat_duration
            player.total_damage = dmg_out_data[2]
            total_damage_out += dmg_out_data[2]
            player.debuff = dmg_out_data[3]
            player.max_one_hit = dmg_out_data[4]
            player.crit_chance = dmg_out_data[5]
            player.total_attacks = dmg_out_data[9]
            player.hull_attacks = dmg_out_data[20]
            player.crit_num = dmg_out_data[11]
            player.misses = dmg_out_data[10]
            dmg_in_data = self.get_player_item(self.damage_in, player_name_handle_id)
            if dmg_in_data is not None:
                dmg_in_data = dmg_in_data.data
                player.deaths = dmg_in_data[8]
                player.total_damage_taken = dmg_in_data[2]
                total_damage_in += dmg_in_data[2]
                player.total_hull_damage_taken = dmg_in_data[15]
                player.total_shield_damage_taken = dmg_in_data[13]
                player.attacks_in_num = dmg_in_data[9]
                total_attacks_in += dmg_in_data[9]
            heal_out_data = self.get_player_item(self.heals_out, player_name_handle_id)
            if heal_out_data is not None:
                heal_out_data = heal_out_data.data
                player.total_heals = heal_out_data[2]
                total_heals += heal_out_data[2]
                player.heal_crit_chance = heal_out_data[8]
                player.heal_crit_num = heal_out_data[10]
                player.heal_num = heal_out_data[9]
            if player.handle in overview_graph_intervals:
                self.create_overview_graphs(player, overview_graph_intervals[player.handle])
            # +++ Currently not used +++
            # build = 'Unknown'
            # for ability_item in player_item._children:
            #     ability_name = ability_item.data[0]
            #     for detection_ability in build_detection_abilities:
            #         if detection_ability in ability_name:
            #             build = Detection.BUILD_DETECTION_ABILITIES[detection_ability]
            #             break
            #     if build != 'Unknown':
            #         break
            # player.build = build
            self.players[player_name_handle_id[0] + player_name_handle_id[1]] = player

        for player in self.players.values():
            try:
                player.heal_share = player.total_heals / total_heals
            except ZeroDivisionError:
                player.heal_share = 0.0
            try:
                player.attacks_in_share = player.attacks_in_num / total_attacks_in
            except ZeroDivisionError:
                player.attacks_in_share = 0.0
            try:
                player.taken_damage_share = player.total_damage_taken / total_damage_in
            except ZeroDivisionError:
                player.taken_damage_share = 0.0
            try:
                player.damage_share = player.total_damage / total_damage_out
            except ZeroDivisionError:
                player.damage_share = 0.0

    def map_is_hive_space(self):
        """
        Checks whether the map is Hive Space by checking for existence of a certain entity.
        """
        for critter in self.damage_in._npc._children:
            if 'Space_Borg_Dreadnought_Hive_Intro' in critter.data.id[0]:
                return True
        return False

    def get_player_item(self, model: TreeModel, name_and_handle: tuple) -> TreeItem | None:
        """
        returns the player item identified by `name_and_handle` from `model`
        """
        for player in model._player._children:
            if player.data[0] == name_and_handle:
                return player

    def detect_map(self):
        """
        Analyzes the entities on the parser to determine:
            - The map type
            - The difficulty of the map
        Requires combat to be fully analyzed
        Fills `self.meta['detection_info']` with information about the detection
        """
        self.map = ''
        critters: dict[str, CritterMeta] = dict()
        for entity in self.damage_in._npc._children:
            entity_name = get_entity_name(entity.data[0][2])
            if entity_name in critters:
                critters[entity_name].add_critter(entity.data[8], entity.data[2])
            else:
                critters[entity_name] = CritterMeta(
                        entity_name, 1, entity.data[8], [entity.data[2]])
        self.critters = critters

        # contains multiple values when multiple entities identify the same map
        map_identificators = critters.keys() & Detection.MAP_IDENTIFIERS_EXISTENCE.keys()
        if len(map_identificators) < 1:
            self.map = 'Combat'
            self.difficulty = None
            self.meta['detection_info'] = [DetectionInfo(False, step='existence')]
            return
        detection_info = list()
        for identificator in map_identificators:
            map_data = Detection.MAP_IDENTIFIERS_EXISTENCE[identificator]
            if map_data['difficulty'] != 'Any':
                self.map = map_data['map']
                self.difficulty = map_data['difficulty']
                self.meta['detection_info'] = [DetectionInfo(
                        True, 'both', (identificator,), step='existence', map=self.map,
                        difficulty=self.difficulty)]
                return
            self.map = map_data['map']
        detection_info.append(
                DetectionInfo(True, 'map', (identificator,), step='existence', map=self.map))

        detected_difficulty = 0
        matched_difficulty = 0
        for diffic, meta in Detection.MAP_DIFFICULTY_ENTITY_DEATH_COUNTS.get(self.map, {}).items():
            detected_difficulty += 1
            if detection_meta := check_difficulty_deaths(meta, critters):
                matched_difficulty += 1
                self.difficulty = diffic
            detection_meta.difficulty = diffic
            detection_info.append(detection_meta)

        # If difficulty wasn't detection  failed in phase 1, don't try phase 2.
        if detected_difficulty and not matched_difficulty:
            self.difficulty = 'Any'
            self.meta['detection_info'] = detection_info
            return

        for diffic, meta in Detection.MAP_DIFFICULTY_ENTITY_HULL_COUNTS.get(self.map, {}).items():
            if detection_meta := check_difficulty_damage(meta, critters):
                self.difficulty = diffic
            detection_meta.difficulty = diffic
            detection_info.append(detection_meta)
        self.meta['detection_info'] = detection_info

    def get_export(self) -> dict:
        analysis_data = dict()
        analysis_data['version'] = '1'
        analysis_data['combat_info'] = {
            'map': self.map,
            'difficulty': self.difficulty,
            'time': self.start_time.timestamp(),
            'log_duration': self.meta['log_duration'],
            'player_duration': self.meta['player_duration'],
        }
        analysis_data['column_names'] = dict()
        analysis_data['column_names']['damage_out'] = list(TREE_HEADER)
        analysis_data['column_names']['damage_out'][0] = 'Ability'
        analysis_data['column_names']['damage_in'] = list(TREE_HEADER)
        analysis_data['column_names']['damage_in'][0] = 'Actor'
        analysis_data['column_names']['heal_out'] = list(TREE_HEADER)
        analysis_data['column_names']['heal_out'][0] = 'Ability'
        analysis_data['column_names']['heal_in'] = list(TREE_HEADER)
        analysis_data['column_names']['heal_in'][0] = 'Actor'
        analysis_data['damage_out'] = analysis_table_export(self.damage_out)
        analysis_data['damage_in'] = analysis_table_export(self.damage_in)
        analysis_data['heals_out'] = analysis_table_export(self.heals_out)
        analysis_data['heals_in'] = analysis_table_export(self.heals_in)
        return analysis_data

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} - Map: {self.map} - Difficulty: {self.difficulty} - "
            f"Datetime: {self.start_time}>"
        )



from collections import namedtuple

import numpy

LogLine = namedtuple(
    'LogLine',
    (
        'timestamp',
        'owner_name',
        'owner_id',
        'source_name',
        'source_id',
        'target_name',
        'target_id',
        'event_name',
        'event_id',
        'type',
        'flags',
        'magnitude',
        'magnitude2'
    )
)


class CritterMeta:
    """
    Represents one npc type in a combat.
    """

    __slots__ = ('name', 'count', 'deaths', 'hull_values')

    def __init__(
            self, name: str, initial_count: int = 0, initial_deaths: int = 0,
            initial_hull_values: list = []):
        """
        Represents one NPC type in a combat

        Parameters:
        - :param name: name of the entity
        - :param initial_count: current count of the entity at time of creation
        - :param initial_hull_values: already determined hull values
        """
        self.name = name
        self.count = initial_count
        self.deaths = initial_deaths
        self.hull_values = initial_hull_values

    def add_critter(self, death: int, hull_value: int):
        """
        Adds an entity to the meta

        Parameters:
        - :param hull_value: amount of hull damage this entity suffered
        """
        self.count += 1
        self.hull_values.append(hull_value)
        self.deaths += death

    def __repr__(self):
        return f'<CritterMeta "Nane: {self.name} Count: {self.count} Deaths: {self.deaths}">'


class DetectionInfo:
    """
    Stores information on the detection process
    """

    __slots__ = (
            'success', 'type', 'identificators', 'target_value', 'retrieved_value', 'step',
            'map', 'difficulty')

    def __init__(
            self, success: bool, type: str = '', identificators: tuple = (), target_value: int = 0,
            retrieved_value: int = 0, step: str = '', map=None, difficulty=None):
        """
        Stores information on the detection process.

        Parameters:
        - :param success: True if this step successfully detected a map
        - :param type: "map"/"difficulty"/"both" depending on what the step detected
        - :param identificators: NPCs that were used to identfy the combat or that the detection \
        failed on
        - :param target_value: value that had to be reached for a successful detection
        - :param retrieved_value: value that was determined from the data
        - :param step: "existence"/"deaths"/"damage" depending on the detection method
        - :param map: detected map
        - :param difficulty: detected difficulty
        """
        self.success = success
        self.type = type
        self.identificators = identificators
        self.target_value = target_value
        self.retrieved_value = retrieved_value
        self.step = step
        self.map = map
        self.difficulty = difficulty

    def __repr__(self):
        return f'<DetectionInfo success={self.success} type={self.type}>'

    def __bool__(self):
        return self.success


class OverviewTableRow:
    '''
    Contains a single row of data
    '''

    __slots__ = (
            'name', 'handle', 'DPS', 'combat_time', 'combat_time_share', 'total_damage', 'debuff',
            'attacks_in_share', 'taken_damage_share', 'damage_share', 'max_one_hit', 'crit_chance',
            'deaths', 'total_heals', 'heal_share', 'heal_crit_chance', 'total_damage_taken',
            'total_hull_damage_taken', 'total_shield_damage_taken', 'total_attacks',
            'hull_attacks', 'attacks_in_num', 'heal_crit_num', 'heal_num', 'crit_num', 'misses',
            'base_damage', 'DMG_graph_data', 'DPS_graph_data', 'graph_time', 'damage_buffer',
            'combat_interval', 'events', 'build')

    def __init__(self, name: str, handle: str):
        self.name: str = name
        self.handle: str = handle
        self.DPS: float = 0.0
        self.combat_time: float = 0.0
        self.combat_time_share: float = 0.0
        self.total_damage: float = 0.0
        self.debuff: float = 0.0
        self.attacks_in_share: float = 0.0
        self.taken_damage_share: float = 0.0
        self.damage_share: float = 0.0
        self.max_one_hit: float = 0.0
        self.crit_chance: float = 0.0
        self.deaths: int = 0
        self.total_heals: float = 0.0
        self.heal_share: float = 0.0
        self.heal_crit_chance: float = 0.0
        self.total_damage_taken: float = 0.0
        self.total_hull_damage_taken: float = 0.0
        self.total_shield_damage_taken: float = 0.0
        self.total_attacks: int = 0
        self.hull_attacks: int = 0
        self.attacks_in_num: int = 0
        self.heal_crit_num: int = 0
        self.heal_num: int = 0
        self.crit_num: int = 0
        self.misses: int = 0

        self.base_damage: float = 0.0
        self.DMG_graph_data: list[float] = list()
        self.DPS_graph_data: list[float] = list()
        self.graph_time: list[float] = list()
        self.damage_buffer: float = 0.0
        self.combat_interval: tuple[float] = None
        self.events: list[str] = list()
        self.build: str = "Unknown"

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.name}{self.handle}>'

    def __len__(self) -> int:
        return 26

    def __getitem__(self, position):
        if position >= 26:
            raise IndexError()
        return getattr(self, self.__slots__[position])

    def to_dict(self):
        return {
            "DPS": self.DPS,
            "name": self.name,
            "deaths": self.deaths,
            "debuff": self.debuff,
            "handle": self.handle,
            "misses": self.misses,
            "crit_num": self.crit_num,
            "heal_num": self.heal_num,
            "heal_share": self.heal_share,
            "combat_time": self.combat_time,
            "crit_chance": self.crit_chance,
            "max_one_hit": self.max_one_hit,
            "total_heals": self.total_heals,
            "damage_share": self.damage_share,
            "hull_attacks": self.hull_attacks,
            "total_damage": self.total_damage,
            "heal_crit_num": self.heal_crit_num,
            "total_attacks": self.total_attacks,
            "attacks_in_num": self.attacks_in_num,
            "attacks_in_share": self.attacks_in_share,
            "heal_crit_chance": self.heal_crit_chance,
            "taken_damage_share": self.taken_damage_share,
            "total_damage_taken": self.total_damage_taken,
            "total_hull_damage_taken": self.total_hull_damage_taken,
            "total_shield_damage_taken": self.total_shield_damage_taken,
            "build": self.build,
        }


class AnalysisTableRow():
    """
    Superclass for damage and heal table rows
    """
    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.name}{self.handle}>'


class DamageTableRow(AnalysisTableRow):
    """
    Contains a single row of data in the analysis table.
    """

    __slots__ = (
            'name', 'handle', 'total_damage', 'max_one_hit', 'kills', 'total_attacks', 'misses',
            'crit_num', 'flank_num', 'total_shield_damage', 'total_hull_damage',
            'total_base_damage', 'combat_time', 'hull_attacks', 'shield_attacks' 'id',
            'combat_start', 'combat_end')

    def __init__(self, name: str, handle: str, id: str):
        """
        Parameters:
        - :param name: name of the entity
        - :param handle: handle of the entity
        - :param id: id of the entity
        """
        # commented attributes represent additional fields in the final result, that are not
        # required here
        self.name: str = name if name else '*'
        self.handle: str = handle
        # self.DPS: float = 0.0
        self.total_damage: float = 0.0
        # self.debuff: float = 0.0
        self.max_one_hit: float = 0.0
        # self.crit_chance: float = 0.0
        # self.accuracy: float = 0.0
        # self.flank_rate: float = 0.0
        self.kills: int = 0
        self.total_attacks: int = 0
        self.misses: int = 0
        self.crit_num: int = 0
        self.flank_num: int = 0
        self.total_shield_damage: float = 0.0
        # self.shield_DPS: float = 0.0
        self.total_hull_damage: float = 0.0
        # self.hull_DPS: float = 0.0
        self.total_base_damage: float = 0.0
        # self.total_base_DPS: float = 0.0
        self.combat_time: float = 0.0
        self.hull_attacks: int = 0
        self.shield_attacks: int = 0

        self.id: str = id
        self.combat_start: float = None
        self.combat_end: float = None


class HealTableRow(AnalysisTableRow):
    """
    Contains a single row of data in the analysis table.
    """
    __slots__ = (
            'name', 'handle', 'total_heal', 'hull_heal', 'shield_heal', 'max_one_heal',
            'heal_ticks', 'critical_heals', 'combat_time', 'hull_heal_ticks', 'shield_heal_ticks',
            'id', 'combat_start', 'combat_end')

    def __init__(self, name: str, handle: str, id: str):
        """
        Parameters:
        - :param name: name of the entity
        - :param handle: handle of the entity
        - :param id: id of the entity
        """

        # commented attributes represent additional fields in the final result, that are not
        # required here
        self.name: str = name if name else '*'
        self.handle: str = handle
        # self.HPS: float = 0.0
        self.total_heal: float = 0.0
        self.hull_heal: float = 0.0
        # self.hull_HPS: float = 0.0
        self.shield_heal: float = 0.0
        # self.shield_HPS: float = 0.0
        self.max_one_heal: float = 0.0
        # self.crit_chance: float = 0.0
        self.heal_ticks: int = 0
        self.critical_heals: int = 0
        self.combat_time: float = 0.0
        self.hull_heal_ticks: int = 0
        self.shield_heal_ticks: int = 0

        self.id: str = id
        self.combat_start: float = None
        self.combat_end: float = None


class TreeItem():
    """
    Item that contains data and children optionally.
    """

    def __init__(self, data: AnalysisTableRow | tuple, parent, parse_duration: int = 0):
        """
        Parameters:
        - :param data: row data
        - :param parent: TreeItem that is the parent of this item
        - :param parse_duration: seconds between the first and last line of the combat, rounded up
        """
        self.data = data
        self.graph_data = numpy.zeros(parse_duration, numpy.float64)
        self.parent: TreeItem = parent
        self._children: list[TreeItem] = list()

    def __repr__(self):
        return f'<{self.__class__.__name__}: data={self.data}>'

    def __iter__(self):
        return iter(self.data)

    def get_child(self, row: int):
        try:
            return self._children[row]
        except IndexError:
            return None

    def append_child(self, item):
        self._children.append(item)

    @property
    def child_count(self):
        return len(self._children)

    @property
    def row(self):
        if self.parent is not None:
            return self.parent._children.index(self)
        return 0

    @property
    def column_count(self):
        return len(self.data)

    def get_data(self, column: int):
        try:
            return self.data[column]
        except IndexError:
            if column == 0:
                return self.data
            return None


class TreeModel():
    """
    Data model that contains a hierachical table.
    """

    def __init__(self, header: tuple[str]):
        self.actor_index = dict()
        self.ability_index: dict[tuple, dict[str, TreeItem]] = dict()
        self.pet_group_index = dict()
        self.pet_index = dict()
        self.target_index = dict()
        self.source_index = dict()
        self._root = TreeItem(header, None)
        self._player = TreeItem(['Player'] + [''] * (len(header) - 1), self._root)
        self._npc = TreeItem(['NPC'] + [''] * (len(header) - 1), self._root)
        self._root.append_child(self._player)
        self._root.append_child(self._npc)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} with columns {self._root.data}>'

    def add_actor(
            self, name: str, handle: str, id: str, row_constructor: AnalysisTableRow,
            player: bool = True) -> TreeItem:
        """
        Adds Player or NPC to model.

        Parameters:
        - :param name: name of the player or NPC
        - :param handle: handle of the player or NPC (NPC handle is an id number)
        - :param id: unique id string of the entity for use in index
        - :param row_constructor: a new object returned by this constructor will hold the row data
        - :param player: True when player is added, False when NPC is added

        :return: added actor
        """
        row = row_constructor(name, handle, id)
        if player:
            actor = TreeItem(row, self._player)
            self._player.append_child(actor)
        else:
            actor = TreeItem(row, self._npc)
            self._npc.append_child(actor)
        self.actor_index[id] = actor
        self.ability_index[id] = dict()
        self.target_index[id] = dict()
        self.source_index[id] = dict()
        return actor

    def add_ability(self, name: str, ability_id: str, actor: TreeItem, actor_id: str) -> TreeItem:
        """
        Adds ability to model. Ability is inserted as child to the given actor.

        Parameters:
        - :param name: name of the ability
        - :param ability_id: unique id string of the ability for use in index
        - :param actor: actor TreeItem that is the parent of the ability

        :return: added ability
        """
        ability = TreeItem(name, actor)
        actor.append_child(ability)
        self.ability_index[actor_id][ability_id] = ability
        self.target_index[actor_id][ability_id] = dict()
        return ability

    def add_pet_group(self, name: str, pet_id: str, actor: TreeItem) -> TreeItem:
        """
        Adds pet group to model. Pet group is inserted as child to the given actor.

        Parameters:
        - :param name: name of the pet group
        - :param pet_id: unique id string of the pet group for use in index
        - :param actor: actor TreeItem that is the parent of the ability

        :return: added pet group
        """
        pet = TreeItem(name, actor)
        actor.append_child(pet)
        self.pet_group_index[pet_id] = pet
        return pet

    def add_pet(self, name: str, pet_id: str, pet_group: TreeItem) -> TreeItem:
        """
        Adds pet to model. Ability is inserted as child to the given pet_group.

        Parameters:
        - :param name: name of the pet
        - :param pet_id: unique id string of the pet for use in index
        - :param pet_group: pet group TreeItem that is the parent of the ability

        :return: added pet
        """
        pet = TreeItem(name, pet_group)
        pet_group.append_child(pet)
        self.pet_index[pet_id] = pet
        self.ability_index[pet_id] = dict()
        self.target_index[pet_id] = dict()
        return pet

    def add_target(
            self, data: AnalysisTableRow, target_id: str, ability: TreeItem, ability_id: str,
            actor_id: str, parse_duration: int) -> TreeItem:
        """
        Adds target to model. target is inserted as child to the given ability.

        Parameters:
        - :param data: row of data for the target. First entry must be the name of the target
        - :param target_id: unique id string of the target for use in index
        - :param ability: ability TreeItem that is the parent of the target
        - :param parse_duration: seconds between the first and last line of the combat, rounded up

        :return: added target
        """
        target = TreeItem(data, ability, parse_duration)
        ability.append_child(target)
        self.target_index[actor_id][ability_id][target_id] = target
        return target

    def add_source_actor(
            self, name: tuple[str], source_id: str, actor: TreeItem, actor_id: str) -> TreeItem:
        """
        Adds source actor to model. Source actor is inserted as child to the given actor. Used for
        incoming damage / heal table.

        Parameters:
        - :param name: name of the source
        - :param source_id: unique id string of the source for use in index
        - :param actor: actor TreeItem that is the parent of the ability

        :return: added source
        """
        source = TreeItem(name, actor)
        actor.append_child(source)
        self.source_index[actor_id][source_id] = source
        self.ability_index[actor_id][source_id] = dict()
        return source

    def add_source_ability(
            self, data: AnalysisTableRow, ability_id: str, source: TreeItem, source_id: str,
            actor_id: str, parse_duration: int) -> TreeItem:
        """
        Adds ability to model. Ability is inserted as child to the given source. Used for incoming
        damage / heal table

        Parameters:
        - :param data: row of data for the target. First entry must be the name of the target
        - :param ability_id: unique id string of the target for use in index
        - :param ability: ability TreeItem that is the parent of the target
        - :param parse_duration: seconds between the first and last line of the combat, rounded up

        :return: added target
        """
        ability = TreeItem(data, source, parse_duration)
        source.append_child(ability)
        self.ability_index[actor_id][source_id][ability_id] = ability
        return ability



from gzip import open as gzip_open
import os
import shutil
from time import time
from typing import Iterable

from .constants import MULTILINE_PATCHES, PATCHES


def format_timestamp(timestamp: str) -> str:
    '''
    Formats timestamp. '24:01:13:04:37:45.7' becomes '24-01-13_04:37:45'
    '''
    return timestamp.replace(':', '-', 2).replace(':', '_', 1).split('.')[0]


def extract_bytes(source_path: str, target_path: str, start_pos: int, end_pos: int):
    """
    Extracts combat from file at `source_path` by copying bytes from `start_pos` (including) up to
    `end_pos` (not including) to a new file at `target_path`

    Parameters:
    - :param source_path: path to source file, must be absolute
    - :param source_path: path to target_file, must be absolute, will overwrite if it already exists
    - :param start_pos: first byte from source file to copy
    - :param end_pos: copies data until this byte, not including it
    """
    if not os.path.isabs(source_path):
        raise AttributeError(f'source_path is not absolute: {source_path}')
    if not os.path.isabs(target_path):
        raise AttributeError(f'target_path is not absolute: {target_path}')
    with open(source_path, 'rb') as source_file:
        if source_file.read(2) == b'\x1f\x8b':
            source_file.close()
            source_file = gzip_open(source_path, 'rb')
        source_file.seek(start_pos)
        extracted_bytes = source_file.read(end_pos - start_pos)
    with open(target_path, 'wb') as target_file:
        target_file.write(extracted_bytes)


def compose_logfile(
            source_path: str, target_path: str, intervals: Iterable[tuple[int, int]],
            templog_folder_path: str):
    """
    Grabs bytes in given `intervals` from `source_path` and writes them to `target_path`.

    Parameters:
    - :param source_path: path to source file, must be absolute
    - :param target_path: path to target file, must be absolute, will overwrite if it already exists
    - :param intervals: iterable with start and end position pairs (half-open interval)
    - :param templog_folder_path: path to folder used for temporary logfiles
    """
    tempfile_path = f'{templog_folder_path}\\{int(time())}'
    with open(source_path, 'rb') as source_file, open(tempfile_path, 'wb') as temp_file:
        if source_file.read(2) == b'\x1f\x8b':
            source_file.close()
            source_file = gzip_open(source_path, 'rb')
        for start_pos, end_pos in intervals:
            source_file.seek(start_pos)
            temp_file.write(source_file.read(end_pos - start_pos))
    shutil.copyfile(tempfile_path, target_path)


def repair_logfile(path: str, templog_folder_path: str) -> str:
    """
    Replace bugged combatlog lines

    Parameters:
    - :param path: logfile to repair
    """
    tempfile_path = f'{templog_folder_path}\\{int(time())}'
    with open(path, 'rb') as log_file, open(tempfile_path, 'wb') as temp_file:
        multiline_progress = 0
        multiline_buffer = b''
        multiline_data = None
        for line in log_file:
            if multiline_progress > 1:
                multiline_progress -= 1
                multiline_buffer += line.strip()
                continue
            elif multiline_progress == 1:
                multiline_progress = 0
                multiline_buffer += line
                temp_file.write(multiline_buffer.replace(multiline_data[0], multiline_data[1]))
                continue
            if line.strip() == b'':
                continue
            for broken_string, fixed_string in PATCHES:
                if broken_string in line:
                    temp_file.write(line.replace(broken_string, fixed_string))
                    break
            else:
                for indentifier, *patch_data in MULTILINE_PATCHES:
                    if indentifier in line:
                        multiline_progress = patch_data[2] - 1
                        multiline_buffer = line.strip()
                        multiline_data = patch_data
                        break
                else:
                    temp_file.write(line)
    try:
        shutil.copyfile(tempfile_path, path)
        res = ''
    except PermissionError:
        res = 'PermissionError'
    return res


def reset_temp_folder(path: str):
    '''
    Deletes and re-creates folder housing temporary log files.
    '''
    if os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            raise FileExistsError(f'Expected path to folder, got "{path}"')
    os.mkdir(path)



"""Combat Detection Methods"""

from .datamodels import LogLine
from .utilities import get_entity_name


class Detection:
    # TODO: Dynamically load as json?
    MAP_IDENTIFIERS_EXISTENCE = {
        "Space_Borg_Battleship_Raidisode_Sibrian_Elite_Initial": {
            "map": "Infected Space",
            "difficulty": "Any",
        },
        "Space_Borg_Dreadnought_Raidisode_Sibrian_Final_Boss": {
            "map": "Infected Space",
            "difficulty": "Any",
        },
        "Mission_Space_Romulan_Colony_Flagship_Lleiset": {
            "map": "Azure Nebula Rescue",
            "difficulty": "Any",
        },
        "Space_Klingon_Dreadnought_Dsc_Sarcophagus": {
            "map": "Battle At The Binary Stars",
            "difficulty": "Any",
        },
        "Event_Procyon_5_Queue_Krenim_Dreadnaught_Annorax": {
            "map": "Battle At Procyon V",
            "difficulty": "Any",
        },
        "Mission_Space_Borg_Queen_Diamond_Brg_Queue_Liberation": {
            "map": "Borg Disconnected",
            "difficulty": "Any",
        },
        "Mission_Starbase_Mirror_Ds9_Mu_Queue": {
            "map": "Counterpoint",
            "difficulty": "Any",
        },
        "Space_Crystalline_Entity_2018": {
            "map": "Crystalline Entity",
            "difficulty": "Any",
        },
        "Event_Ico_Qonos_Space_Herald_Dreadnaught": {
            "map": "Gateway To Grethor",
            "difficulty": "Any",
        },
        "Mission_Space_Federation_Science_Herald_Sphere": {
            "map": "Herald Sphere",
            "difficulty": "Any",
        },
        "Msn_Dsc_Priors_System_Tfo_Orbital_Platform_1_Fed_Dsc": {
            "map": "Operation Riposte",
            "difficulty": "Any",
        },
        "Space_Borg_Dreadnought_R02": {
            "map": "Cure Found",
            "difficulty": "Any",
        },
        "Space_Klingon_Tos_X3_Battlecruiser": {
            "map": "Days Of Doom",
            "difficulty": "Any",
        },
        "Msn_Luk_Colony_Dranuur_Queue_System_Upgradeable_Satellite": {
            "map": "Dranuur Gauntlet",
            "difficulty": "Any",
        },
        "Space_Borg_Dreadnought_Raidisode_Khitomer_Intro_Boss": {
            "map": "Khitomer Space",
            "difficulty": "Any",
        },
        "Mission_Spire_Space_Voth_Frigate": {
            "map": "Storming The Spire",
            "difficulty": "Any",
        },
        "Space_Drantzuli_Alpha_Battleship": {
            "map": "Swarm",
            "difficulty": "Any",
        },
        "Mission_Beta_Lankal_Destructible_Reactor": {
            "map": "To Hell With Honor",
            "difficulty": "Any",
        },
        "Space_Federation_Dreadnought_Jupiter_Class_Carrier": {
            "map": "Gravity Kills",
            "difficulty": "Any",
        },
        "Msn_Luk_Hypermass_Queue_System_Tzk_Protomatter_Facility": {
            "map": "Gravity Kills",
            "difficulty": "Any",
        },
        "Space_Borg_Dreadnought_Hive_Intro": {
            "map": "Hive Space",
            "difficulty": "Any",
        },
        "Ground_Federation_Capt_Mirror_Runabout_Tfo": {
            "map": "Operation Wolf",
            "difficulty": "Normal",
        },
        "Bluegills_Ground_Boss": {
            "map": "Bug Hunt",
            "difficulty": "Any",
        },
        "Msn_Edren_Queue_Ground_Gorn_Lt_Tos_Range_Rock": {
            "map": "Miner Instabilities",
            "difficulty": "Any",
        },
        "Msn_Ground_Capt_Mirror_Janeway_Boss_Unkillable": {
            "map": "Jupiter Station Showdown",
            "difficulty": "Any",
        },
        "Mission_Event_Tholian_Invasion_Ext_Boss": {
            "map": "Nukara Prime: Transdimensional Tactics",
            "difficulty": "Any",
        },
        "Space_Borg_Dreadnought_Wolf359": {
            "map": "Battle of Wolf 359",
            "difficulty": "Any",
        },
        "Snowman_Q_Boss_Msn_Snowglobe": {
            "map": "Winter Invasion",
            "difficulty": "Normal",
        },
        "Dlt_Vaadwaur_Stf_System_Dreadnought_Boss": {
            "map": "Battle of Korfez",
            "difficulty": "Elite"
        }
    }

    # Detect maps based on # of entities
    # Note for these dicts: Difficulty needs to be ordered from low to high.
    # e.g. Elite needs to be placed AFTER Advanced else the map detection may
    # always detect the lower difficulty when the higher difficulty may match.
    MAP_DIFFICULTY_ENTITY_DEATH_COUNTS = {
        "Infected Space": {
            "Advanced": {
                "Space_Borg_Battleship_Raidisode": 5,
                "Space_Borg_Cruiser_Raidisode": 6,
                "Mission_Borgraid1_Transwarp_02": 1,
                "Space_Borg_Dreadnought_Raidisode_Sibrian_Final_Boss": 1,
            },
            "Elite": {
                "Space_Borg_Battleship_Raidisode_Sibrian_Elite_Initial": 2,
                "Space_Borg_Dreadnought_Raidisode_Sibrian_Initial_Boss": 1,
                "Space_Borg_Cruiser_Raidisode_Sibrian_Elite_Initial": 4,
                "Space_Borg_Battleship_Raidisode": 2,
                "Mission_Borgraid1_Transwarp_02": 1,
                "Space_Borg_Dreadnought_Raidisode_Sibrian_Final_Boss": 1,
            },
        },
        "Cure Found": {
            "Advanced": {
                "Space_Borg_Battleship_Raidisode_Cure": 3,
                "Mission_Cure_Healer_Mini_Trans_02": 18,
                "Space_Borg_Cruiser_Raidisode_Cure": 3,
                "Space_Borg_Cruiser_Raidisode": 2,
                "Space_Borg_Dreadnought_R02": 1,
                "Space_Klingon_Raider_Pet_Borg_Carrier_Advanced": 0,
            },
            "Elite": {
                "Space_Borg_Battleship_Raidisode_Cure": 3,
                "Mission_Cure_Healer_Mini_Trans_02": 18,
                "Space_Borg_Cruiser_Raidisode_Cure": 3,
                "Space_Borg_Cruiser_Raidisode": 2,
                "Space_Borg_Dreadnought_R02": 1,
                "Space_Klingon_Fighter_Pet_Borg_Elite": 0,
            },
        },
        "Khitomer Space": {
            "Advanced": {
                "Space_Borg_Dreadnought_Raidisode_Khitomer_Intro_Boss": 1,
                "Mission_Raidisode03_Donatra_Borg_Scimitar": 1,
                "Mission_Borgraid1_Transwarp_02": 2,
                "Space_Borg_Battleship_Raidisode": 4,
                "Mission_Borgraid1_Comm_Array": 4,
                "Space_Borg_Dreadnought_Raidisode": 0,
            },
            "Elite": {
                "Space_Borg_Dreadnought_Raidisode_Khitomer_Intro_Boss": 1,
                "Mission_Raidisode03_Donatra_Borg_Scimitar": 1,
                "Mission_Borgraid1_Transwarp_02": 2,
                "Space_Borg_Battleship_Raidisode": 4,
                "Mission_Borgraid1_Comm_Array": 4,
                "Space_Borg_Dreadnought_Raidisode": 4,
            },
        },
        "Hive Space": {
            "Advanced": {
                "Mission_Space_Borg_Queen_Diamond": 1,
                "Mission_Space_Borg_Battleship_Queen_2_0f_2": 1,
                "Mission_Space_Borg_Battleship_Queen_1_0f_2": 1,
            },
            "Elite": {
                "Mission_Space_Borg_Queen_Diamond": 1,
                "Mission_Space_Borg_Battleship_Queen_2_0f_2": 1,
                "Mission_Space_Borg_Battleship_Queen_1_0f_2": 1,
            },
        },
        "Bug Hunt": {
            "Elite": {
                "Msn_Dlt_Bluegill_Hunt_Queue_Ground_Ens": 3,
                "Bluegills_Ground_Cdr": 26,
                "Bluegills_Ground_Capt": 1,
                "Bluegills_Ground_Boss": 1,
            },
        },
        "Jupiter Station Showdown": {
            "Elite": {
                "Msn_Assimilated_Fed_Odyssey_Ground_Borg_Ens_Melee": 27,
                "Msn_Assimilated_Fed_Odyssey_Ground_Borg_Lt_Range": 17,
                "Msn_Assimilated_Fed_Odyssey_Ground_Borg_Cdr_Melee": 2,
            }
        },
        "Miner Instabilities": {
            "Elite": {
                "Ground_Nakuhl_Capt_Range_Male": 1,
            }
        },
        "Battle of Wolf 359": {
            "Elite": {
                "Space_Borg_Cruiser_Wolf359": 3,
            }
        },
    }

    # Detect maps based on # hull damage taken.
    MAP_DIFFICULTY_ENTITY_HULL_COUNTS = {
        "Hive Space": {
            "Advanced": {
                "Space_Borg_Cruiser_Hive_Intro1": 461582,
                "Space_Borg_Cruiser_Hive_Intro2": 461582,
                "Space_Borg_Battleship_Hive_Intro": 576977,
                "Space_Borg_Dreadnought_Hive_Intro": 1707034,
            },
            "Elite": {
                "Space_Borg_Cruiser_Hive_Intro1": 2165239,
                "Space_Borg_Cruiser_Hive_Intro2": 2165239,
                "Space_Borg_Battleship_Hive_Intro": 2706549,
                "Space_Borg_Dreadnought_Hive_Intro": 8007542,
            },
        },
        "Jupiter Station Showdown": {
            "Elite": {
                "Msn_Assimilated_Fed_Odyssey_Ground_Borg_Ens_Melee": 2605,
                "Msn_Assimilated_Fed_Odyssey_Ground_Borg_Lt_Range": 3439,
            }
        },
        "Bug Hunt": {
            # It's very easy to overkill on ground.
            "Elite": {
                # "Bluegills_Ground_Cdr": 9444,
                # "Bluegills_Ground_Ens": 3191,
                # "Bluegills_Ground_Lt": 4986,
                # "Bluegills_Ground_Capt": 32567,
                "Bluegills_Ground_Boss": 449432,
            }
        },
        "Miner Instabilities": {
            # It's very easy to overkill on ground.
            "Elite": {
                "Ground_Romulan_Tos_Cdr_Range": 6513,
                # "Ground_Romulan_Tos_Lt_Range": 3439,
                # "Ground_Romulan_Tos_Ens_Range": 2605,
                # "Ground_Nakuhl_Cdr_Range_Male": 6513,
                # "Ground_Nakuhl_Cdr_Range_Female": 6513,
                # "Ground_Nakuhl_Lt_Range_Male": 3439,
                # "Ground_Nakuhl_Lt_Range_Female": 3439,
                # "Ground_Nakuhl_Ens_Melee": 2605,
                # "Ground_Nakuhl_Ens_Range": 2605,
                "Ground_Nakuhl_Capt_Range_Male": 20843,
            }
        },
        "Battle of Wolf 359": {
            "Elite": {
                "Space_Borg_Turret_Medium_Plasma_Torpedo_Wolf359": 2081960,
                "Space_Borg_Turret_Medium_Plasma_Beam_Wolf359": 2081960,
                "Space_Borg_Turret_Medium_Tractor_Beam_Wolf359": 2081960,
                "Space_Borg_Wolf359_Escape_Pod_Tractor_Beam": 2081960,  # Assumption
                "Space_Borg_Frigate_Wolf359": 2081960,
                "Space_Borg_Cruiser_Wolf359": 0,
            }
        },
    }

    BUILD_DETECTION_ABILITIES = {
        # DEW
        "Surgical Strikes III": "Surgical Strikes",
        "Reroute Reserves to Weapons": "Reroute Reserves to Weapons",
        "Exceed Rated Limits": "Exceed Rated Limits",
        "Rapid Fire III": "Cannons: Rapid Fire",
        "Scatter Volley III": "Cannons: Scatter Volley",
        "Overload III": "Beams: Overload",
        "Fire at Will III": "Beams: Fire At Will",
        # Kinetic
        "Isolytic Tear": "Kinetic",
        # EPG
        "Electrified Anomalies": "Exotic",
        "Deteriorating Secondary Deflector": "Exotic",
        "Gravity Well III": "Exotic",
        # Support
        "Greater Than The Sum": "Support",
        # Lower Rank abilities that may be commonly used.
        "Rapid Fire II": "Cannons: Rapid Fire",
        "Scatter Volley II": "Cannons: Scatter Volley",
        "Fire at Will II": "Beams: Fire At Will",
        # Classify these last
        "Thalaron Pulse": "Thalaron Pulse",
    }

    @staticmethod
    def detect_line(line: LogLine) -> tuple:
        """
        Do a very shallow map detect based on a log line only
        taking NPCs into consideration.

        return: a tuple in the form of (Map, Difficulty)
        """

        # Note: Doing a string split is slightly faster than using a regex.
        # re.search(r"C\[.* (?P<name>.*)]", line.target_id)

        entity = get_entity_name(line.target_id)
        if entity is None:
            return "Combat", None

        entry = Detection.MAP_IDENTIFIERS_EXISTENCE.get(entity)
        if entry:
            return entry["map"], entry["difficulty"]

        return "Combat", None



from .datamodels import TreeModel


def analysis_table_export(table_model: TreeModel) -> dict:
    """
    Returns only player combat data from table, only considering one hierachy level below the
    player nodes.

    :param table_model: model containing the table/tree
    """
    data = dict()
    for player in table_model._player._children:
        player_data = [ability.data for ability in player._children]
        data[player.data[0][2]] = player_data
    return data
